from _typeshed import Self
from typing import Any, Iterable, MutableSequence, Sequence, Text, overload

class UserString(Sequence[UserString]):
    data: unicode
    def __init__(self, seq: object) -> None: ...
    def __int__(self) -> int: ...
    def __long__(self) -> long: ...
    def __float__(self) -> float: ...
    def __complex__(self) -> complex: ...
    def __hash__(self) -> int: ...
    def __len__(self) -> int: ...
    @overload
    def __getitem__(self: Self, i: int) -> Self: ...
    @overload
    def __getitem__(self: Self, s: slice) -> Self: ...
    def __add__(self: Self, other: Any) -> Self: ...
    def __radd__(self: Self, other: Any) -> Self: ...
    def __mul__(self: Self, other: int) -> Self: ...
    def __rmul__(self: Self, other: int) -> Self: ...
    def __mod__(self: Self, args: Any) -> Self: ...
    def capitalize(self: Self) -> Self: ...
    def center(self: Self, width: int, *args: Any) -> Self: ...
    def count(self, sub: int, start: int = ..., end: int = ...) -> int: ...
    def decode(self: Self, encoding: str | None = ..., errors: str | None = ...) -> Self: ...
    def encode(self: Self, encoding: str | None = ..., errors: str | None = ...) -> Self: ...
    def endswith(self, suffix: Text | tuple[Text, ...], start: int | None = ..., end: int | None = ...) -> bool: ...
    def expandtabs(self: Self, tabsize: int = ...) -> Self: ...
    def find(self, sub: Text, start: int = ..., end: int = ...) -> int: ...
    def index(self, sub: Text, start: int = ..., end: int = ...) -> int: ...
    def isalpha(self) -> bool: ...
    def isalnum(self) -> bool: ...
    def isdecimal(self) -> bool: ...
    def isdigit(self) -> bool: ...
    def islower(self) -> bool: ...
    def isnumeric(self) -> bool: ...
    def isspace(self) -> bool: ...
    def istitle(self) -> bool: ...
    def isupper(self) -> bool: ...
    def join(self, seq: Iterable[Text]) -> Text: ...
    def ljust(self: Self, width: int, *args: Any) -> Self: ...
    def lower(self: Self) -> Self: ...
    def lstrip(self: Self, chars: Text | None = ...) -> Self: ...
    def partition(self, sep: Text) -> tuple[Text, Text, Text]: ...
    def replace(self: Self, old: Text, new: Text, maxsplit: int = ...) -> Self: ...
    def rfind(self, sub: Text, start: int = ..., end: int = ...) -> int: ...
    def rindex(self, sub: Text, start: int = ..., end: int = ...) -> int: ...
    def rjust(self: Self, width: int, *args: Any) -> Self: ...
    def rpartition(self, sep: Text) -> tuple[Text, Text, Text]: ...
    def rstrip(self: Self, chars: Text | None = ...) -> Self: ...
    def split(self, sep: Text | None = ..., maxsplit: int = ...) -> list[Text]: ...
    def rsplit(self, sep: Text | None = ..., maxsplit: int = ...) -> list[Text]: ...
    def splitlines(self, keepends: int = ...) -> list[Text]: ...
    def startswith(self, prefix: Text | tuple[Text, ...], start: int | None = ..., end: int | None = ...) -> bool: ...
    def strip(self: Self, chars: Text | None = ...) -> Self: ...
    def swapcase(self: Self) -> Self: ...
    def title(self: Self) -> Self: ...
    def translate(self: Self, *args: Any) -> Self: ...
    def upper(self: Self) -> Self: ...
    def zfill(self: Self, width: int) -> Self: ...

class MutableString(UserString, MutableSequence[MutableString]):
    @overload
    def __getitem__(self: Self, i: int) -> Self: ...
    @overload
    def __getitem__(self: Self, s: slice) -> Self: ...
    def __setitem__(self, index: int | slice, sub: Any) -> None: ...
    def __delitem__(self, index: int | slice) -> None: ...
    def immutable(self) -> UserString: ...
    def __iadd__(self: Self, other: Any) -> Self: ...
    def __imul__(self: Self, n: int) -> Self: ...
    def insert(self, index: int, value: Any) -> None: ...



from typing import Any, TypeVar

from pygments.formatter import Formatter

_T = TypeVar("_T", str, bytes)

class IRCFormatter(Formatter[_T]):
    name: str
    aliases: Any
    filenames: Any
    darkbg: Any
    colorscheme: Any
    linenos: Any
    def format_unencoded(self, tokensource, outfile) -> None: ...



from typing import Any

from .base import BaseEndpoint as BaseEndpoint

log: Any

class SignatureOnlyEndpoint(BaseEndpoint):
    def validate_request(self, uri, http_method: str = ..., body: Any | None = ..., headers: Any | None = ...): ...



from _typeshed import SupportsLessThanT
from typing import Sequence, Text, overload
from typing_extensions import Literal

# All overloads can return empty string. Ideally, Literal[""] would be a valid
# Iterable[T], so that Union[List[T], Literal[""]] could be used as a return
# type. But because this only works when T is str, we need Sequence[T] instead.
@overload
def commonprefix(m: Sequence[str]) -> str | Literal[""]: ...  # type: ignore[misc]
@overload
def commonprefix(m: Sequence[Text]) -> Text: ...
@overload
def commonprefix(m: Sequence[list[SupportsLessThanT]]) -> Sequence[SupportsLessThanT]: ...
@overload
def commonprefix(m: Sequence[tuple[SupportsLessThanT, ...]]) -> Sequence[SupportsLessThanT]: ...
def exists(path: Text) -> bool: ...
def getsize(filename: Text) -> int: ...
def isfile(path: Text) -> bool: ...
def isdir(s: Text) -> bool: ...

# These return float if os.stat_float_times() == True,
# but int is a subclass of float.
def getatime(filename: Text) -> float: ...
def getmtime(filename: Text) -> float: ...
def getctime(filename: Text) -> float: ...



from _typeshed import SupportsLessThan
from typing import Callable, Iterable, TypeVar

_T = TypeVar("_T")

def cmp_lt(x, y) -> bool: ...
def heappush(heap: list[_T], item: _T) -> None: ...
def heappop(heap: list[_T]) -> _T: ...
def heappushpop(heap: list[_T], item: _T) -> _T: ...
def heapify(x: list[_T]) -> None: ...
def heapreplace(heap: list[_T], item: _T) -> _T: ...
def merge(*iterables: Iterable[_T]) -> Iterable[_T]: ...
def nlargest(n: int, iterable: Iterable[_T], key: Callable[[_T], SupportsLessThan] | None = ...) -> list[_T]: ...
def nsmallest(n: int, iterable: Iterable[_T], key: Callable[[_T], SupportsLessThan] | None = ...) -> list[_T]: ...
def _heapify_max(__x: list[_T]) -> None: ...  # undocumented



import argparse
import ast
from typing import Any, Generic, Iterable, Iterator, TypeVar

FLAKE8_ERROR = tuple[int, int, str, type[Any]]
TConfig = TypeVar("TConfig")  # noqa: Y001  # Name of the TypeVar matches the name at runtime

class Error:
    code: str
    message: str
    lineno: int
    col_offset: int
    def __init__(self, lineno: int, col_offset: int, **kwargs: Any) -> None: ...
    @classmethod
    def formatted_message(cls, **kwargs: Any) -> str: ...

class Visitor(ast.NodeVisitor, Generic[TConfig]):
    errors: list[Error]
    def __init__(self, config: TConfig | None = ...) -> None: ...
    @property
    def config(self) -> TConfig: ...
    def error_from_node(self, error: type[Error], node: ast.AST, **kwargs: Any) -> None: ...

class Plugin(Generic[TConfig]):
    name: str
    version: str
    visitors: list[type[Visitor[TConfig]]]
    config: TConfig
    def __init__(self, tree: ast.AST) -> None: ...
    def run(self) -> Iterable[FLAKE8_ERROR]: ...
    @classmethod
    def parse_options(cls, option_manager: Any, options: argparse.Namespace, args: list[str]) -> None: ...
    @classmethod
    def parse_options_to_config(cls, option_manager: Any, options: argparse.Namespace, args: list[str]) -> TConfig | None: ...
    @classmethod
    def test_config(cls, config: TConfig) -> Iterator[None]: ...



from typing import Any, Pattern

_DeclarativeMeta = Any  # SQLAlchemy is not part of typeshed

def should_set_tablename(cls: type) -> bool: ...

camelcase_re: Pattern[str]

def camel_to_snake_case(name: str) -> str: ...

class NameMetaMixin(type):
    def __init__(cls, name, bases, d) -> None: ...
    def __table_cls__(cls, *args, **kwargs): ...

class BindMetaMixin(type):
    def __init__(cls, name, bases, d) -> None: ...

class DefaultMeta(NameMetaMixin, BindMetaMixin, _DeclarativeMeta): ...

class Model:
    query_class: Any | None
    query: Any | None



from typing import Any

from .icalendar import VCalendar2_0

class HCalendar(VCalendar2_0):
    name: str
    @classmethod
    def serialize(cls, obj, buf: Any | None = ..., lineLength: Any | None = ..., validate: bool = ...): ...



# syntax error doesn't get detected by 2to3

def f(*): pass


# Gets parsed as:
# Node(funcdef,
#      [Leaf(1, 'def'), Leaf(1, 'f'),
#       Node(parameters, [Leaf(7, '('), Leaf(16, '*'), Leaf(8, ')')]),
#       Leaf(11, ':'),
#       Node(simple_stmt, [Leaf(1, 'pass')])])



from collections.abc import Generator
from typing import Any

from selenium.common.exceptions import (
    InvalidArgumentException as InvalidArgumentException,
    NoSuchCookieException as NoSuchCookieException,
    WebDriverException as WebDriverException,
)
from selenium.webdriver.common.by import By as By
from selenium.webdriver.common.html5.application_cache import ApplicationCache as ApplicationCache

from .command import Command as Command
from .errorhandler import ErrorHandler as ErrorHandler
from .file_detector import FileDetector as FileDetector, LocalFileDetector as LocalFileDetector
from .mobile import Mobile as Mobile
from .remote_connection import RemoteConnection as RemoteConnection
from .switch_to import SwitchTo as SwitchTo
from .webelement import WebElement as WebElement

class WebDriver:
    command_executor: Any
    session_id: Any
    capabilities: Any
    error_handler: Any
    def __init__(
        self,
        command_executor: str | RemoteConnection = ...,
        desired_capabilities: Any | None = ...,
        browser_profile: Any | None = ...,
        proxy: Any | None = ...,
        keep_alive: bool = ...,
        file_detector: Any | None = ...,
        options: Any | None = ...,
    ) -> None: ...
    def __enter__(self): ...
    def __exit__(self, *args) -> None: ...
    def file_detector_context(self, file_detector_class, *args, **kwargs) -> Generator[None, None, None]: ...
    @property
    def mobile(self): ...
    @property
    def name(self) -> str: ...
    def start_client(self) -> None: ...
    def stop_client(self) -> None: ...
    w3c: Any
    def start_session(self, capabilities, browser_profile: Any | None = ...) -> None: ...
    def create_web_element(self, element_id) -> WebElement: ...
    def execute(self, driver_command, params: Any | None = ...): ...
    def get(self, url) -> None: ...
    @property
    def title(self) -> str: ...
    def find_element_by_id(self, id_) -> WebElement: ...
    def find_elements_by_id(self, id_) -> list[WebElement]: ...
    def find_element_by_xpath(self, xpath) -> WebElement: ...
    def find_elements_by_xpath(self, xpath) -> list[WebElement]: ...
    def find_element_by_link_text(self, link_text) -> WebElement: ...
    def find_elements_by_link_text(self, text) -> list[WebElement]: ...
    def find_element_by_partial_link_text(self, link_text) -> WebElement: ...
    def find_elements_by_partial_link_text(self, link_text) -> list[WebElement]: ...
    def find_element_by_name(self, name) -> WebElement: ...
    def find_elements_by_name(self, name) -> list[WebElement]: ...
    def find_element_by_tag_name(self, name) -> WebElement: ...
    def find_elements_by_tag_name(self, name) -> list[WebElement]: ...
    def find_element_by_class_name(self, name) -> WebElement: ...
    def find_elements_by_class_name(self, name) -> list[WebElement]: ...
    def find_element_by_css_selector(self, css_selector) -> WebElement: ...
    def find_elements_by_css_selector(self, css_selector) -> list[WebElement]: ...
    def execute_script(self, script, *args): ...
    def execute_async_script(self, script, *args): ...
    @property
    def current_url(self) -> str: ...
    @property
    def page_source(self) -> str: ...
    def close(self) -> None: ...
    def quit(self) -> None: ...
    @property
    def current_window_handle(self) -> str: ...
    @property
    def window_handles(self) -> list[str]: ...
    def maximize_window(self) -> None: ...
    def fullscreen_window(self) -> None: ...
    def minimize_window(self) -> None: ...
    @property
    def switch_to(self) -> SwitchTo: ...
    def switch_to_active_element(self): ...
    def switch_to_window(self, window_name) -> None: ...
    def switch_to_frame(self, frame_reference) -> None: ...
    def switch_to_default_content(self) -> None: ...
    def switch_to_alert(self): ...
    def back(self) -> None: ...
    def forward(self) -> None: ...
    def refresh(self) -> None: ...
    def get_cookies(self): ...
    def get_cookie(self, name): ...
    def delete_cookie(self, name) -> None: ...
    def delete_all_cookies(self) -> None: ...
    def add_cookie(self, cookie_dict) -> None: ...
    def implicitly_wait(self, time_to_wait) -> None: ...
    def set_script_timeout(self, time_to_wait) -> None: ...
    def set_page_load_timeout(self, time_to_wait) -> None: ...
    def find_element(self, by=..., value: Any | None = ...) -> WebElement: ...
    def find_elements(self, by=..., value: Any | None = ...) -> list[WebElement]: ...
    @property
    def desired_capabilities(self): ...
    def get_screenshot_as_file(self, filename) -> bool: ...
    def save_screenshot(self, filename) -> bool: ...
    def get_screenshot_as_png(self) -> bytes: ...
    def get_screenshot_as_base64(self) -> str: ...
    def set_window_size(self, width, height, windowHandle: str = ...) -> None: ...
    def get_window_size(self, windowHandle: str = ...): ...
    def set_window_position(self, x, y, windowHandle: str = ...): ...
    def get_window_position(self, windowHandle: str = ...): ...
    def get_window_rect(self): ...
    def set_window_rect(self, x: Any | None = ..., y: Any | None = ..., width: Any | None = ..., height: Any | None = ...): ...
    @property
    def file_detector(self): ...
    @file_detector.setter
    def file_detector(self, detector) -> None: ...
    @property
    def orientation(self): ...
    @orientation.setter
    def orientation(self, value) -> None: ...
    @property
    def application_cache(self): ...
    @property
    def log_types(self): ...
    def get_log(self, log_type): ...



from typing import Any

from docutils.nodes import Node, document

class Transform:
    def __init__(self, document: document, startnode: Node | None = ...): ...
    def __getattr__(self, __name: str) -> Any: ...  # incomplete

class Transformer:
    def __init__(self, document: document): ...
    def add_transform(self, transform_class: type[Transform], priority: int | None = ..., **kwargs) -> None: ...
    def __getattr__(self, __name: str) -> Any: ...  # incomplete

def __getattr__(name: str) -> Any: ...  # incomplete



class Options:
    KEY: str
    def __init__(self) -> None: ...
    @property
    def binary_location(self): ...
    @binary_location.setter
    def binary_location(self, value) -> None: ...
    @property
    def capabilities(self): ...
    def set_capability(self, name, value) -> None: ...
    @property
    def debugger_address(self): ...
    @debugger_address.setter
    def debugger_address(self, value) -> None: ...
    @property
    def arguments(self): ...
    def add_argument(self, argument) -> None: ...
    @property
    def extensions(self): ...
    def add_extension(self, extension) -> None: ...
    def add_encoded_extension(self, extension) -> None: ...
    @property
    def experimental_options(self): ...
    def add_experimental_option(self, name, value) -> None: ...
    @property
    def headless(self): ...
    @headless.setter
    def headless(self, value) -> None: ...
    def set_headless(self, headless: bool = ...) -> None: ...
    def to_capabilities(self): ...



import array
import threading
import weakref
from queue import Queue as Queue
from typing import Any, Callable, Iterable, Mapping, Sequence
from typing_extensions import Literal

from .connection import Pipe as Pipe

__all__ = [
    "Process",
    "current_process",
    "active_children",
    "freeze_support",
    "Lock",
    "RLock",
    "Semaphore",
    "BoundedSemaphore",
    "Condition",
    "Event",
    "Barrier",
    "Queue",
    "Manager",
    "Pipe",
    "Pool",
    "JoinableQueue",
]

JoinableQueue = Queue
Barrier = threading.Barrier
BoundedSemaphore = threading.BoundedSemaphore
Condition = threading.Condition
Event = threading.Event
Lock = threading.Lock
RLock = threading.RLock
Semaphore = threading.Semaphore

class DummyProcess(threading.Thread):
    _children: weakref.WeakKeyDictionary[Any, Any]
    _parent: threading.Thread
    _pid: None
    _start_called: int
    @property
    def exitcode(self) -> Literal[0] | None: ...
    def __init__(
        self,
        group: Any = ...,
        target: Callable[..., Any] | None = ...,
        name: str | None = ...,
        args: Iterable[Any] = ...,
        kwargs: Mapping[str, Any] = ...,
    ) -> None: ...

Process = DummyProcess

class Namespace:
    def __init__(self, **kwds: Any) -> None: ...
    def __getattr__(self, __name: str) -> Any: ...
    def __setattr__(self, __name: str, __value: Any) -> None: ...

class Value:
    _typecode: Any
    _value: Any
    value: Any
    def __init__(self, typecode: Any, value: Any, lock: Any = ...) -> None: ...

def Array(typecode: Any, sequence: Sequence[Any], lock: Any = ...) -> array.array[Any]: ...
def Manager() -> Any: ...
def Pool(processes: int | None = ..., initializer: Callable[..., Any] | None = ..., initargs: Iterable[Any] = ...) -> Any: ...
def active_children() -> list[Any]: ...
def current_process() -> threading.Thread: ...
def freeze_support() -> None: ...
def shutdown() -> None: ...



import abc
from distutils.cmd import Command as _Command
from typing import Any

po_file_read_mode: Any

def listify_value(arg, split: Any | None = ...): ...

class Command(_Command, metaclass=abc.ABCMeta):
    as_args: Any
    multiple_value_options: Any
    boolean_options: Any
    option_aliases: Any
    option_choices: Any
    log: Any
    distribution: Any
    verbose: bool
    force: Any
    help: int
    finalized: int
    def __init__(self, dist: Any | None = ...) -> None: ...

class compile_catalog(Command):
    description: str
    user_options: Any
    boolean_options: Any
    domain: str
    directory: Any
    input_file: Any
    output_file: Any
    locale: Any
    use_fuzzy: bool
    statistics: bool
    def initialize_options(self) -> None: ...
    def finalize_options(self) -> None: ...
    def run(self): ...

class extract_messages(Command):
    description: str
    user_options: Any
    boolean_options: Any
    as_args: str
    multiple_value_options: Any
    option_aliases: Any
    option_choices: Any
    charset: str
    keywords: Any
    no_default_keywords: bool
    mapping_file: Any
    no_location: bool
    add_location: Any
    omit_header: bool
    output_file: Any
    input_dirs: Any
    input_paths: Any
    width: Any
    no_wrap: bool
    sort_output: bool
    sort_by_file: bool
    msgid_bugs_address: Any
    copyright_holder: Any
    project: Any
    version: Any
    add_comments: Any
    strip_comments: bool
    include_lineno: bool
    def initialize_options(self) -> None: ...
    def finalize_options(self) -> None: ...
    def run(self) -> None: ...

def check_message_extractors(dist, name, value) -> None: ...

class init_catalog(Command):
    description: str
    user_options: Any
    boolean_options: Any
    output_dir: Any
    output_file: Any
    input_file: Any
    locale: Any
    domain: str
    no_wrap: bool
    width: Any
    def initialize_options(self) -> None: ...
    def finalize_options(self) -> None: ...
    def run(self) -> None: ...

class update_catalog(Command):
    description: str
    user_options: Any
    boolean_options: Any
    domain: str
    input_file: Any
    output_dir: Any
    output_file: Any
    omit_header: bool
    locale: Any
    width: Any
    no_wrap: bool
    ignore_obsolete: bool
    init_missing: bool
    no_fuzzy_matching: bool
    update_header_comment: bool
    previous: bool
    def initialize_options(self) -> None: ...
    def finalize_options(self) -> None: ...
    def run(self) -> None: ...

class CommandLineInterface:
    usage: str
    version: Any
    commands: Any
    command_classes: Any
    log: Any
    parser: Any
    def run(self, argv: Any | None = ...): ...

def main(): ...
def parse_mapping(fileobj, filename: Any | None = ...): ...
def parse_keywords(strings=...): ...



"""Representation of modules."""

import dataclasses
import os
from typing import Sequence


@dataclasses.dataclass(eq=True, frozen=True)
class Module:
  """Inferred information about a module.

  Attributes:
    path: The path to the module, e.g., foo/.
    target: The filename relative to the path, e.g., bar/baz.py.
    name: The module name, e.g., bar.baz.
    kind: The module kind: Builtin, Direct, Local, or System.
      See https://github.com/google/importlab/blob/main/importlab/resolve.py.
    full_path: The full path to the module (path + target).
  """

  path: str
  target: str
  name: str
  kind: str = "Local"

  @property
  def full_path(self):
    return os.path.join(self.path, self.target)


def infer_module(filename, pythonpath):
  """Convert a filename to a module relative to pythonpath.

  This method tries to deduce the module name from the pythonpath and the
  filename. This will not always be possible. (It depends on the filename
  starting with an entry in the pythonpath.)

  Args:
    filename: The filename of a Python file. E.g. "foo/bar/baz.py".
    pythonpath: The path Python uses to search for modules.

  Returns:
    A Module object.
  """
  # We want '' in our lookup path, but we don't want it for prefix tests.
  for path in filter(bool, pythonpath):
    if not path.endswith(os.sep):
      path += os.sep
    if filename.startswith(path):
      filename = filename[len(path):]
      break
  else:
    # We have not found filename relative to anywhere in pythonpath.
    path = ""
  return Module(path, filename, path_to_module_name(filename))


def get_module_name(filename, pythonpath):
  """Get the module name, or None if we can't determine it."""
  if filename:
    filename = os.path.normpath(filename)
    # Keep path '' as is; infer_module will handle it.
    pythonpath = [path and os.path.normpath(path) for path in pythonpath]
    return infer_module(filename, pythonpath).name


def path_to_module_name(filename):
  """Converts a filename into a dotted module name."""
  if os.path.dirname(filename).startswith(os.pardir):
    # Don't try to infer a module name for filenames starting with ../
    return None
  filename, ext = os.path.splitext(filename)
  if ext and not ext.startswith(".py"):
    # If there is no extension, convert "foo/bar" to "foo.bar", since we use
    # that in our imports_info map. If there is an extension, it needs to be
    # a python source or stub file, so ".py*" should cover all the cases.
    return None
  module_name = filename.replace(os.path.sep, ".")
  # strip __init__ suffix
  module_name, _, _ = module_name.partition(".__init__")
  return module_name


def strip_init_suffix(parts: Sequence[str]):
  return parts[:-1] if parts and parts[-1] == "__init__" else parts


def get_absolute_name(prefix, relative_name):
  """Joins a dotted-name prefix and a relative name.

  Args:
    prefix: A dotted name, e.g. foo.bar.baz
    relative_name: A dotted name with possibly some leading dots, e.g. ..x.y

  Returns:
    The relative name appended to the prefix, after going up one level for each
      leading dot.
      e.g. foo.bar.baz + ..hello.world -> foo.bar.hello.world
    None if the relative name has too many leading dots.
  """
  path = prefix.split(".") if prefix else []
  name = relative_name.lstrip(".")
  ndots = len(relative_name) - len(name)
  if ndots > len(path):
    return None
  prefix = "".join([p + "." for p in path[:len(path) + 1 - ndots]])
  return prefix + name


def get_package_name(module_name, is_package=False):
  """Figure out a package name for a module."""
  if module_name is None:
    return ""
  parts = module_name.split(".")
  if not is_package:
    parts = parts[:-1]
  return ".".join(parts)


def get_all_prefixes(module_name):
  """Return all the prefixes of a module name.

  e.g. x.y.z => x, x.y, x.y.z

  Args:
    module_name: module name

  Returns:
    List of prefixes
  """
  parts = module_name.split(".")
  name = parts[0]
  out = [name]
  for part in parts[1:]:
    name = ".".join([name, part])
    out.append(name)
  return out



from typing import Any, AnyStr, Iterable, Mapping, Text

from . import compat, cookies, exceptions, structures

OrderedDict = compat.OrderedDict
RequestsCookieJar = cookies.RequestsCookieJar
cookiejar_from_dict = cookies.cookiejar_from_dict
CaseInsensitiveDict = structures.CaseInsensitiveDict
InvalidURL = exceptions.InvalidURL

NETRC_FILES: Any
DEFAULT_CA_BUNDLE_PATH: Any
DEFAULT_PORTS: Any

def dict_to_sequence(d): ...
def super_len(o): ...
def get_netrc_auth(url, raise_errors: bool = ...): ...
def guess_filename(obj): ...
def extract_zipped_paths(path): ...
def from_key_val_list(value): ...
def to_key_val_list(value): ...
def parse_list_header(value): ...
def parse_dict_header(value): ...
def unquote_header_value(value, is_filename=...): ...
def dict_from_cookiejar(cj): ...
def add_dict_to_cookiejar(cj, cookie_dict): ...
def get_encodings_from_content(content): ...
def get_encoding_from_headers(headers): ...
def stream_decode_response_unicode(iterator, r): ...
def iter_slices(string, slice_length): ...
def get_unicode_from_response(r): ...

UNRESERVED_SET: Any

def unquote_unreserved(uri): ...
def requote_uri(uri): ...
def address_in_network(ip, net): ...
def dotted_netmask(mask): ...
def is_ipv4_address(string_ip): ...
def is_valid_cidr(string_network): ...
def set_environ(env_name, value): ...
def should_bypass_proxies(url, no_proxy: Iterable[AnyStr] | None) -> bool: ...
def get_environ_proxies(url, no_proxy: Iterable[AnyStr] | None = ...) -> dict[Any, Any]: ...
def select_proxy(url: Text, proxies: Mapping[Any, Any] | None): ...
def default_user_agent(name=...): ...
def default_headers(): ...
def parse_header_links(value): ...
def guess_json_utf(data): ...
def prepend_scheme_if_needed(url, new_scheme): ...
def get_auth_from_url(url): ...
def to_native_string(string, encoding=...): ...
def urldefragauth(url): ...
def rewind_body(prepared_request): ...
def check_header_validity(header: tuple[AnyStr, AnyStr]) -> None: ...



from typing import Any, BinaryIO, Callable, Protocol, Text, overload

class _ReadableBinary(Protocol):
    def tell(self) -> int: ...
    def read(self, size: int) -> bytes: ...
    def seek(self, offset: int) -> Any: ...

_File = Text | _ReadableBinary

@overload
def what(file: _File, h: None = ...) -> str | None: ...
@overload
def what(file: Any, h: bytes) -> str | None: ...

tests: list[Callable[[bytes, BinaryIO | None], str | None]]



# Types to support PEP 3333 (WSGI)
#
# This module doesn't exist at runtime and neither do the types defined in this
# file. They are provided for type checking purposes.

from sys import _OptExcInfo
from typing import Any, Callable, Iterable, Protocol, Text

class StartResponse(Protocol):
    def __call__(
        self, status: str, headers: list[tuple[str, str]], exc_info: _OptExcInfo | None = ...
    ) -> Callable[[bytes], Any]: ...

WSGIEnvironment = dict[Text, Any]
WSGIApplication = Callable[[WSGIEnvironment, StartResponse], Iterable[bytes]]

# WSGI input streams per PEP 3333
class InputStream(Protocol):
    def read(self, size: int = ...) -> bytes: ...
    def readline(self, size: int = ...) -> bytes: ...
    def readlines(self, hint: int = ...) -> list[bytes]: ...
    def __iter__(self) -> Iterable[bytes]: ...

# WSGI error streams per PEP 3333
class ErrorStream(Protocol):
    def flush(self) -> None: ...
    def write(self, s: str) -> None: ...
    def writelines(self, seq: list[str]) -> None: ...

class _Readable(Protocol):
    def read(self, size: int = ...) -> bytes: ...

# Optional file wrapper in wsgi.file_wrapper
class FileWrapper(Protocol):
    def __call__(self, file: _Readable, block_size: int = ...) -> Iterable[bytes]: ...



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

class ErrorBars(Serialisable):
    tagname: str
    errDir: Any
    direction: Any
    errBarType: Any
    style: Any
    errValType: Any
    size: Any
    noEndCap: Any
    plus: Any
    minus: Any
    val: Any
    spPr: Any
    graphicalProperties: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self,
        errDir: Any | None = ...,
        errBarType: str = ...,
        errValType: str = ...,
        noEndCap: Any | None = ...,
        plus: Any | None = ...,
        minus: Any | None = ...,
        val: Any | None = ...,
        spPr: Any | None = ...,
        extLst: Any | None = ...,
    ) -> None: ...



from html.parser import HTMLParser
from typing import Any

from bs4.builder import HTMLTreeBuilder

class HTMLParseError(Exception): ...

class BeautifulSoupHTMLParser(HTMLParser):
    IGNORE: str
    REPLACE: str
    on_duplicate_attribute: Any
    already_closed_empty_element: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def error(self, msg: str) -> None: ...
    def handle_startendtag(self, name, attrs) -> None: ...
    def handle_starttag(self, name, attrs, handle_empty_element: bool = ...) -> None: ...
    def handle_endtag(self, name, check_already_closed: bool = ...) -> None: ...
    def handle_data(self, data) -> None: ...
    def handle_charref(self, name) -> None: ...
    def handle_entityref(self, name) -> None: ...
    def handle_comment(self, data) -> None: ...
    def handle_decl(self, data) -> None: ...
    def unknown_decl(self, data) -> None: ...
    def handle_pi(self, data) -> None: ...

class HTMLParserTreeBuilder(HTMLTreeBuilder):
    is_xml: bool
    picklable: bool
    NAME: Any
    features: Any
    TRACKS_LINE_NUMBERS: bool
    parser_args: Any
    def __init__(self, parser_args: Any | None = ..., parser_kwargs: Any | None = ..., **kwargs) -> None: ...
    def prepare_markup(
        self,
        markup,
        user_specified_encoding: Any | None = ...,
        document_declared_encoding: Any | None = ...,
        exclude_encodings: Any | None = ...,
    ) -> None: ...
    def feed(self, markup) -> None: ...



from typing import Any

from jmespath import parser as parser
from jmespath.visitor import Options as Options

def compile(expression): ...
def search(expression, data, options: Any | None = ...): ...



from typing import Any

from .collection import Collection as Collection
from .config import Config as Config
from .context import Context as Context, MockContext as MockContext
from .exceptions import (
    AmbiguousEnvVar as AmbiguousEnvVar,
    AuthFailure as AuthFailure,
    CollectionNotFound as CollectionNotFound,
    CommandTimedOut as CommandTimedOut,
    Exit as Exit,
    Failure as Failure,
    ParseError as ParseError,
    PlatformError as PlatformError,
    ResponseNotAccepted as ResponseNotAccepted,
    SubprocessPipeError as SubprocessPipeError,
    ThreadException as ThreadException,
    UncastableEnvVar as UncastableEnvVar,
    UnexpectedExit as UnexpectedExit,
    UnknownFileType as UnknownFileType,
    UnpicklableConfigMember as UnpicklableConfigMember,
    WatcherError as WatcherError,
)
from .executor import Executor as Executor
from .loader import FilesystemLoader as FilesystemLoader
from .parser import Argument as Argument, Parser as Parser, ParserContext as ParserContext, ParseResult as ParseResult
from .program import Program as Program
from .runners import Local as Local, Promise as Promise, Result as Result, Runner as Runner
from .tasks import Call as Call, Task as Task, call as call, task as task
from .terminals import pty_size as pty_size
from .watchers import FailingResponder as FailingResponder, Responder as Responder, StreamWatcher as StreamWatcher

__version_info__: tuple[int, int, int]
__version__: str

def run(command: str, **kwargs: Any) -> Result: ...
def sudo(command: str, **kwargs: Any) -> Result: ...



import builtins
from collections.abc import Callable, Iterable, Iterator, Mapping, Sequence
from datetime import datetime, timedelta
from typing import Any, Generic, TypeVar, overload
from typing_extensions import Literal

from ..client import _CommandOptions, _Key, _Value

_ScoreCastFuncReturn = TypeVar("_ScoreCastFuncReturn")
_StrType = TypeVar("_StrType", bound=str | bytes)

class ACLCommands(Generic[_StrType]):
    def acl_cat(self, category: str | None = ..., **kwargs: _CommandOptions) -> list[str]: ...
    def acl_deluser(self, *username: str, **kwargs: _CommandOptions) -> int: ...
    def acl_genpass(self, bits: int | None = ..., **kwargs: _CommandOptions) -> str: ...
    def acl_getuser(self, username: str, **kwargs: _CommandOptions) -> Any | None: ...
    def acl_help(self, **kwargs: _CommandOptions): ...
    def acl_list(self, **kwargs: _CommandOptions) -> list[str]: ...
    def acl_log(self, count: int | None = ..., **kwargs: _CommandOptions): ...
    def acl_log_reset(self, **kwargs: _CommandOptions): ...
    def acl_load(self, **kwargs: _CommandOptions) -> bool: ...
    def acl_save(self, **kwargs: _CommandOptions): ...
    def acl_setuser(
        self,
        username: str,
        enabled: bool = ...,
        nopass: bool = ...,
        passwords: Sequence[str] | None = ...,
        hashed_passwords: Sequence[str] | None = ...,
        categories: Sequence[str] | None = ...,
        commands: Sequence[str] | None = ...,
        keys: Sequence[str] | None = ...,
        reset: bool = ...,
        reset_keys: bool = ...,
        reset_passwords: bool = ...,
        **kwargs: _CommandOptions,
    ) -> bool: ...
    def acl_users(self, **kwargs: _CommandOptions) -> list[str]: ...
    def acl_whoami(self, **kwargs: _CommandOptions) -> str: ...

class ManagementCommands:
    def bgrewriteaof(self, **kwargs: _CommandOptions): ...
    def bgsave(self, schedule: bool = ..., **kwargs: _CommandOptions): ...
    def role(self): ...
    def client_kill(self, address: str, **kwargs: _CommandOptions) -> bool: ...
    def client_kill_filter(
        self,
        _id: Any | None = ...,
        _type: Any | None = ...,
        addr: Any | None = ...,
        skipme: Any | None = ...,
        laddr: Any | None = ...,
        user: Any | None = ...,
        **kwargs: _CommandOptions,
    ): ...
    def client_info(self, **kwargs: _CommandOptions): ...
    def client_list(
        self, _type: str | None = ..., client_id: list[str] = ..., **kwargs: _CommandOptions
    ) -> list[dict[str, str]]: ...
    def client_getname(self, **kwargs: _CommandOptions) -> str | None: ...
    def client_getredir(self, **kwargs: _CommandOptions): ...
    def client_reply(self, reply, **kwargs: _CommandOptions): ...
    def client_id(self, **kwargs: _CommandOptions) -> int: ...
    def client_tracking_on(
        self, clientid: Any | None = ..., prefix=..., bcast: bool = ..., optin: bool = ..., optout: bool = ..., noloop: bool = ...
    ): ...
    def client_tracking_off(
        self, clientid: Any | None = ..., prefix=..., bcast: bool = ..., optin: bool = ..., optout: bool = ..., noloop: bool = ...
    ): ...
    def client_tracking(
        self,
        on: bool = ...,
        clientid: Any | None = ...,
        prefix=...,
        bcast: bool = ...,
        optin: bool = ...,
        optout: bool = ...,
        noloop: bool = ...,
        **kwargs: _CommandOptions,
    ): ...
    def client_trackinginfo(self, **kwargs: _CommandOptions): ...
    def client_setname(self, name: str, **kwargs: _CommandOptions) -> bool: ...
    def client_unblock(self, client_id, error: bool = ..., **kwargs: _CommandOptions): ...
    def client_pause(self, timeout, all: bool = ..., **kwargs: _CommandOptions): ...
    def client_unpause(self, **kwargs: _CommandOptions): ...
    def command(self, **kwargs: _CommandOptions): ...
    def command_info(self, **kwargs: _CommandOptions): ...
    def command_count(self, **kwargs: _CommandOptions): ...
    def config_get(self, pattern: str = ..., **kwargs: _CommandOptions): ...
    def config_set(self, name, value, **kwargs: _CommandOptions): ...
    def config_resetstat(self, **kwargs: _CommandOptions): ...
    def config_rewrite(self, **kwargs: _CommandOptions): ...
    def dbsize(self, **kwargs: _CommandOptions) -> int: ...
    def debug_object(self, key, **kwargs: _CommandOptions): ...
    def debug_segfault(self, **kwargs: _CommandOptions): ...
    def echo(self, value: _Value, **kwargs: _CommandOptions) -> bytes: ...
    def flushall(self, asynchronous: bool = ..., **kwargs: _CommandOptions) -> bool: ...
    def flushdb(self, asynchronous: bool = ..., **kwargs: _CommandOptions) -> bool: ...
    def sync(self): ...
    def psync(self, replicationid, offset): ...
    def swapdb(self, first, second, **kwargs: _CommandOptions): ...
    def select(self, index, **kwargs: _CommandOptions): ...
    def info(self, section: _Key | None = ..., **kwargs: _CommandOptions) -> Mapping[str, Any]: ...
    def lastsave(self, **kwargs: _CommandOptions): ...
    def lolwut(self, *version_numbers: _Value, **kwargs: _CommandOptions) -> bytes: ...
    def reset(self) -> None: ...
    def migrate(
        self,
        host,
        port,
        keys,
        destination_db,
        timeout,
        copy: bool = ...,
        replace: bool = ...,
        auth: Any | None = ...,
        **kwargs: _CommandOptions,
    ): ...
    def object(self, infotype, key, **kwargs: _CommandOptions): ...
    def memory_doctor(self, **kwargs: _CommandOptions): ...
    def memory_help(self, **kwargs: _CommandOptions): ...
    def memory_stats(self, **kwargs: _CommandOptions) -> dict[str, Any]: ...
    def memory_malloc_stats(self, **kwargs: _CommandOptions): ...
    def memory_usage(self, key, samples: Any | None = ..., **kwargs: _CommandOptions): ...
    def memory_purge(self, **kwargs: _CommandOptions): ...
    def ping(self, **kwargs: _CommandOptions) -> bool: ...
    def quit(self, **kwargs: _CommandOptions): ...
    def replicaof(self, *args, **kwargs: _CommandOptions): ...
    def save(self, **kwargs: _CommandOptions) -> bool: ...
    def shutdown(self, save: bool = ..., nosave: bool = ..., **kwargs: _CommandOptions) -> None: ...
    def slaveof(self, host: Any | None = ..., port: Any | None = ..., **kwargs: _CommandOptions): ...
    def slowlog_get(self, num: Any | None = ..., **kwargs: _CommandOptions): ...
    def slowlog_len(self, **kwargs: _CommandOptions): ...
    def slowlog_reset(self, **kwargs: _CommandOptions): ...
    def time(self, **kwargs: _CommandOptions): ...
    def wait(self, num_replicas, timeout, **kwargs: _CommandOptions): ...

class BasicKeyCommands(Generic[_StrType]):
    def append(self, key, value): ...
    def bitcount(self, key: _Key, start: int | None = ..., end: int | None = ...) -> int: ...
    def bitfield(self, key, default_overflow: Any | None = ...): ...
    def bitop(self, operation, dest, *keys): ...
    def bitpos(self, key, bit, start=..., end=...): ...
    def copy(self, source, destination, destination_db: Any | None = ..., replace: bool = ...): ...
    def decr(self, name, amount: int = ...) -> int: ...
    def decrby(self, name, amount: int = ...) -> int: ...
    def delete(self, *names: _Key) -> int: ...
    def __delitem__(self, name: _Key) -> None: ...
    def dump(self, name: _Key) -> _StrType | None: ...
    def exists(self, *names: _Key) -> int: ...
    __contains__ = exists
    def expire(self, name: _Key, time: int | timedelta) -> bool: ...
    def expireat(self, name, when): ...
    def get(self, name: _Key) -> _StrType | None: ...
    def getdel(self, name: _Key) -> _StrType | None: ...
    def getex(
        self,
        name,
        ex: Any | None = ...,
        px: Any | None = ...,
        exat: Any | None = ...,
        pxat: Any | None = ...,
        persist: bool = ...,
    ): ...
    def __getitem__(self, name: str): ...
    def getbit(self, name: _Key, offset: int) -> int: ...
    def getrange(self, key, start, end): ...
    def getset(self, name, value) -> _StrType | None: ...
    def incr(self, name: _Key, amount: int = ...) -> int: ...
    def incrby(self, name: _Key, amount: int = ...) -> int: ...
    def incrbyfloat(self, name: _Key, amount: float = ...) -> float: ...
    def keys(self, pattern: _Key = ..., **kwargs: _CommandOptions) -> list[_StrType]: ...
    def lmove(
        self, first_list: _Key, second_list: _Key, src: Literal["LEFT", "RIGHT"] = ..., dest: Literal["LEFT", "RIGHT"] = ...
    ) -> _Value: ...
    def blmove(
        self,
        first_list: _Key,
        second_list: _Key,
        timeout: float,
        src: Literal["LEFT", "RIGHT"] = ...,
        dest: Literal["LEFT", "RIGHT"] = ...,
    ) -> _Value | None: ...
    def mget(self, keys: _Key | Iterable[_Key], *args: _Key) -> list[_StrType | None]: ...
    def mset(self, mapping: Mapping[_Key, _Value]) -> Literal[True]: ...
    def msetnx(self, mapping: Mapping[_Key, _Value]) -> bool: ...
    def move(self, name: _Key, db: int) -> bool: ...
    def persist(self, name: _Key) -> bool: ...
    def pexpire(self, name: _Key, time: int | timedelta) -> Literal[1, 0]: ...
    def pexpireat(self, name: _Key, when: int | datetime) -> Literal[1, 0]: ...
    def psetex(self, name, time_ms, value): ...
    def pttl(self, name: _Key) -> int: ...
    def hrandfield(self, key, count: Any | None = ..., withvalues: bool = ...): ...
    def randomkey(self, **kwargs: _CommandOptions): ...
    def rename(self, src, dst): ...
    def renamenx(self, src, dst): ...
    def restore(
        self, name, ttl, value, replace: bool = ..., absttl: bool = ..., idletime: Any | None = ..., frequency: Any | None = ...
    ): ...
    def set(
        self,
        name: _Key,
        value: _Value,
        ex: None | int | timedelta = ...,
        px: None | int | timedelta = ...,
        nx: bool = ...,
        xx: bool = ...,
        keepttl: bool = ...,
        get: bool = ...,
        exat: Any | None = ...,
        pxat: Any | None = ...,
    ) -> bool | None: ...
    def __setitem__(self, name, value) -> None: ...
    def setbit(self, name: _Key, offset: int, value: int) -> int: ...
    def setex(self, name: _Key, time: int | timedelta, value: _Value) -> bool: ...
    def setnx(self, name: _Key, value: _Value) -> bool: ...
    def setrange(self, name, offset, value): ...
    def stralgo(
        self,
        algo,
        value1,
        value2,
        specific_argument: str = ...,
        len: bool = ...,
        idx: bool = ...,
        minmatchlen: Any | None = ...,
        withmatchlen: bool = ...,
        **kwargs: _CommandOptions,
    ): ...
    def strlen(self, name): ...
    def substr(self, name, start, end: int = ...): ...
    def touch(self, *args): ...
    def ttl(self, name: _Key) -> int: ...
    def type(self, name): ...
    def watch(self, *names): ...
    def unwatch(self): ...
    def unlink(self, *names: _Key) -> int: ...

class ListCommands(Generic[_StrType]):
    @overload
    def blpop(self, keys: _Value | Iterable[_Value], timeout: Literal[0] | None = ...) -> tuple[_StrType, _StrType]: ...
    @overload
    def blpop(self, keys: _Value | Iterable[_Value], timeout: float) -> tuple[_StrType, _StrType] | None: ...
    @overload
    def brpop(self, keys: _Value | Iterable[_Value], timeout: Literal[0] | None = ...) -> tuple[_StrType, _StrType]: ...
    @overload
    def brpop(self, keys: _Value | Iterable[_Value], timeout: float) -> tuple[_StrType, _StrType] | None: ...
    def brpoplpush(self, src, dst, timeout: int | None = ...): ...
    def lindex(self, name: _Key, index: int) -> _StrType | None: ...
    def linsert(
        self, name: _Key, where: Literal["BEFORE", "AFTER", "before", "after"], refvalue: _Value, value: _Value
    ) -> int: ...
    def llen(self, name: _Key) -> int: ...
    def lpop(self, name, count: int | None = ...): ...
    def lpush(self, name: _Value, *values: _Value) -> int: ...
    def lpushx(self, name, value): ...
    def lrange(self, name: _Key, start: int, end: int) -> list[_StrType]: ...
    def lrem(self, name: _Key, count: int, value: _Value) -> int: ...
    def lset(self, name: _Key, index: int, value: _Value) -> bool: ...
    def ltrim(self, name: _Key, start: int, end: int) -> bool: ...
    def rpop(self, name, count: int | None = ...): ...
    def rpoplpush(self, src, dst): ...
    def rpush(self, name: _Value, *values: _Value) -> int: ...
    def rpushx(self, name, value): ...
    def lpos(self, name, value, rank: Any | None = ..., count: Any | None = ..., maxlen: Any | None = ...): ...
    @overload
    def sort(
        self,
        name: _Key,
        start: int | None = ...,
        num: int | None = ...,
        by: _Key | None = ...,
        get: _Key | Sequence[_Key] | None = ...,
        desc: bool = ...,
        alpha: bool = ...,
        store: None = ...,
        groups: bool = ...,
    ) -> list[_StrType]: ...
    @overload
    def sort(
        self,
        name: _Key,
        start: int | None = ...,
        num: int | None = ...,
        by: _Key | None = ...,
        get: _Key | Sequence[_Key] | None = ...,
        desc: bool = ...,
        alpha: bool = ...,
        *,
        store: _Key,
        groups: bool = ...,
    ) -> int: ...
    @overload
    def sort(
        self,
        name: _Key,
        start: int | None,
        num: int | None,
        by: _Key | None,
        get: _Key | Sequence[_Key] | None,
        desc: bool,
        alpha: bool,
        store: _Key,
        groups: bool = ...,
    ) -> int: ...

class ScanCommands(Generic[_StrType]):
    def scan(
        self,
        cursor: int = ...,
        match: _Key | None = ...,
        count: int | None = ...,
        _type: str | None = ...,
        **kwargs: _CommandOptions,
    ) -> tuple[int, list[_StrType]]: ...
    def scan_iter(
        self, match: _Key | None = ..., count: int | None = ..., _type: str | None = ..., **kwargs: _CommandOptions
    ) -> Iterator[_StrType]: ...
    def sscan(
        self, name: _Key, cursor: int = ..., match: _Key | None = ..., count: int | None = ...
    ) -> tuple[int, list[_StrType]]: ...
    def sscan_iter(self, name: _Key, match: _Key | None = ..., count: int | None = ...) -> Iterator[_StrType]: ...
    def hscan(
        self, name: _Key, cursor: int = ..., match: _Key | None = ..., count: int | None = ...
    ) -> tuple[int, dict[_StrType, _StrType]]: ...
    def hscan_iter(
        self, name: _Key, match: _Key | None = ..., count: int | None = ...
    ) -> Iterator[tuple[_StrType, _StrType]]: ...
    def zscan(
        self,
        name: _Key,
        cursor: int = ...,
        match: _Key | None = ...,
        count: int | None = ...,
        score_cast_func: Callable[[_StrType], _ScoreCastFuncReturn] = ...,
    ) -> tuple[int, list[tuple[_StrType, _ScoreCastFuncReturn]]]: ...
    def zscan_iter(
        self,
        name: _Key,
        match: _Key | None = ...,
        count: int | None = ...,
        score_cast_func: Callable[[_StrType], _ScoreCastFuncReturn] = ...,
    ) -> Iterator[tuple[_StrType, _ScoreCastFuncReturn]]: ...

class SetCommands(Generic[_StrType]):
    def sadd(self, name: _Key, *values: _Value) -> int: ...
    def scard(self, name: _Key) -> int: ...
    def sdiff(self, keys: _Key | Iterable[_Key], *args: _Key) -> builtins.set[_Value]: ...
    def sdiffstore(self, dest: _Key, keys: _Key | Iterable[_Key], *args: _Key) -> int: ...
    def sinter(self, keys: _Key | Iterable[_Key], *args: _Key) -> builtins.set[_Value]: ...
    def sinterstore(self, dest: _Key, keys: _Key | Iterable[_Key], *args: _Key) -> int: ...
    def sismember(self, name: _Key, value: _Value) -> bool: ...
    def smembers(self, name: _Key) -> builtins.set[_StrType]: ...
    def smismember(self, name, values, *args): ...
    def smove(self, src: _Key, dst: _Key, value: _Value) -> bool: ...
    @overload
    def spop(self, name: _Key, count: None = ...) -> _Value | None: ...
    @overload
    def spop(self, name: _Key, count: int) -> list[_Value]: ...
    @overload
    def srandmember(self, name: _Key, number: None = ...) -> _Value | None: ...
    @overload
    def srandmember(self, name: _Key, number: int) -> list[_Value]: ...
    def srem(self, name: _Key, *values: _Value) -> int: ...
    def sunion(self, keys: _Key | Iterable[_Key], *args: _Key) -> builtins.set[_Value]: ...
    def sunionstore(self, dest: _Key, keys: _Key | Iterable[_Key], *args: _Key) -> int: ...

class StreamCommands:
    def xack(self, name, groupname, *ids): ...
    def xadd(
        self,
        name,
        fields,
        id: str = ...,
        maxlen=...,
        approximate: bool = ...,
        nomkstream: bool = ...,
        minid: Any | None = ...,
        limit: Any | None = ...,
    ): ...
    def xautoclaim(
        self, name, groupname, consumername, min_idle_time, start_id: int = ..., count: Any | None = ..., justid: bool = ...
    ): ...
    def xclaim(
        self, name, groupname, consumername, min_idle_time, message_ids, idle=..., time=..., retrycount=..., force=..., justid=...
    ): ...
    def xdel(self, name, *ids): ...
    def xgroup_create(self, name, groupname, id: str = ..., mkstream: bool = ...): ...
    def xgroup_delconsumer(self, name, groupname, consumername): ...
    def xgroup_destroy(self, name, groupname): ...
    def xgroup_createconsumer(self, name, groupname, consumername): ...
    def xgroup_setid(self, name, groupname, id): ...
    def xinfo_consumers(self, name, groupname): ...
    def xinfo_groups(self, name): ...
    def xinfo_stream(self, name, full: bool = ...): ...
    def xlen(self, name: _Key) -> int: ...
    def xpending(self, name, groupname): ...
    def xpending_range(
        self,
        name,
        groupname,
        idle: Any | None = ...,
        min: Any | None = ...,
        max: Any | None = ...,
        count: int | None = ...,
        consumername: Any | None = ...,
    ): ...
    def xrange(self, name, min: str = ..., max: str = ..., count: Any | None = ...): ...
    def xread(self, streams, count: Any | None = ..., block: Any | None = ...): ...
    def xreadgroup(
        self, groupname, consumername, streams, count: Any | None = ..., block: Any | None = ..., noack: bool = ...
    ): ...
    def xrevrange(self, name, max: str = ..., min: str = ..., count: Any | None = ...): ...
    def xtrim(
        self, name, maxlen: Any | None = ..., approximate: bool = ..., minid: Any | None = ..., limit: Any | None = ...
    ): ...

class SortedSetCommands(Generic[_StrType]):
    def zadd(
        self,
        name: _Key,
        mapping: Mapping[_Key, _Value],
        nx: bool = ...,
        xx: bool = ...,
        ch: bool = ...,
        incr: bool = ...,
        gt: Any | None = ...,
        lt: Any | None = ...,
    ) -> int: ...
    def zcard(self, name: _Key) -> int: ...
    def zcount(self, name: _Key, min: _Value, max: _Value) -> int: ...
    def zdiff(self, keys, withscores: bool = ...): ...
    def zdiffstore(self, dest, keys): ...
    def zincrby(self, name: _Key, amount: float, value: _Value) -> float: ...
    def zinter(self, keys, aggregate: Any | None = ..., withscores: bool = ...): ...
    def zinterstore(self, dest: _Key, keys: Iterable[_Key], aggregate: Literal["SUM", "MIN", "MAX"] | None = ...) -> int: ...
    def zlexcount(self, name: _Key, min: _Value, max: _Value) -> int: ...
    def zpopmax(self, name: _Key, count: int | None = ...) -> list[tuple[_StrType, float]]: ...
    def zpopmin(self, name: _Key, count: int | None = ...) -> list[tuple[_StrType, float]]: ...
    def zrandmember(self, key, count: Any | None = ..., withscores: bool = ...): ...
    @overload
    def bzpopmax(self, keys: _Key | Iterable[_Key], timeout: Literal[0] = ...) -> tuple[_StrType, _StrType, float]: ...
    @overload
    def bzpopmax(self, keys: _Key | Iterable[_Key], timeout: float) -> tuple[_StrType, _StrType, float] | None: ...
    @overload
    def bzpopmin(self, keys: _Key | Iterable[_Key], timeout: Literal[0] = ...) -> tuple[_StrType, _StrType, float]: ...
    @overload
    def bzpopmin(self, keys: _Key | Iterable[_Key], timeout: float) -> tuple[_StrType, _StrType, float] | None: ...
    @overload
    def zrange(
        self,
        name: _Key,
        start: int,
        end: int,
        desc: bool,
        withscores: Literal[True],
        score_cast_func: Callable[[_StrType], _ScoreCastFuncReturn] = ...,
        byscore: bool = ...,
        bylex: bool = ...,
        offset: int | None = ...,
        num: int | None = ...,
    ) -> list[tuple[_StrType, _ScoreCastFuncReturn]]: ...
    @overload
    def zrange(
        self,
        name: _Key,
        start: int,
        end: int,
        desc: bool = ...,
        *,
        withscores: Literal[True],
        score_cast_func: Callable[[_StrType], _ScoreCastFuncReturn] = ...,
        byscore: bool = ...,
        bylex: bool = ...,
        offset: int | None = ...,
        num: int | None = ...,
    ) -> list[tuple[_StrType, _ScoreCastFuncReturn]]: ...
    @overload
    def zrange(
        self,
        name: _Key,
        start: int,
        end: int,
        desc: bool = ...,
        withscores: bool = ...,
        score_cast_func: Callable[[_StrType], Any] = ...,
        byscore: bool = ...,
        bylex: bool = ...,
        offset: int | None = ...,
        num: int | None = ...,
    ) -> list[_StrType]: ...
    @overload
    def zrevrange(
        self,
        name: _Key,
        start: int,
        end: int,
        withscores: Literal[True],
        score_cast_func: Callable[[_StrType], _ScoreCastFuncReturn] = ...,
    ) -> list[tuple[_StrType, _ScoreCastFuncReturn]]: ...
    @overload
    def zrevrange(
        self, name: _Key, start: int, end: int, withscores: bool = ..., score_cast_func: Callable[[Any], Any] = ...
    ) -> list[_StrType]: ...
    def zrangestore(
        self,
        dest,
        name,
        start,
        end,
        byscore: bool = ...,
        bylex: bool = ...,
        desc: bool = ...,
        offset: Any | None = ...,
        num: Any | None = ...,
    ): ...
    def zrangebylex(
        self, name: _Key, min: _Value, max: _Value, start: int | None = ..., num: int | None = ...
    ) -> list[_StrType]: ...
    def zrevrangebylex(
        self, name: _Key, max: _Value, min: _Value, start: int | None = ..., num: int | None = ...
    ) -> list[_StrType]: ...
    @overload
    def zrangebyscore(
        self,
        name: _Key,
        min: _Value,
        max: _Value,
        start: int | None = ...,
        num: int | None = ...,
        *,
        withscores: Literal[True],
        score_cast_func: Callable[[_StrType], _ScoreCastFuncReturn] = ...,
    ) -> list[tuple[_StrType, _ScoreCastFuncReturn]]: ...
    @overload
    def zrangebyscore(
        self,
        name: _Key,
        min: _Value,
        max: _Value,
        start: int | None = ...,
        num: int | None = ...,
        withscores: bool = ...,
        score_cast_func: Callable[[_StrType], Any] = ...,
    ) -> list[_StrType]: ...
    @overload
    def zrevrangebyscore(
        self,
        name: _Key,
        max: _Value,
        min: _Value,
        start: int | None = ...,
        num: int | None = ...,
        *,
        withscores: Literal[True],
        score_cast_func: Callable[[_StrType], _ScoreCastFuncReturn] = ...,
    ) -> list[tuple[_StrType, _ScoreCastFuncReturn]]: ...
    @overload
    def zrevrangebyscore(
        self,
        name: _Key,
        max: _Value,
        min: _Value,
        start: int | None = ...,
        num: int | None = ...,
        withscores: bool = ...,
        score_cast_func: Callable[[_StrType], Any] = ...,
    ) -> list[_StrType]: ...
    def zrank(self, name: _Key, value: _Value) -> int | None: ...
    def zrem(self, name: _Key, *values: _Value) -> int: ...
    def zremrangebylex(self, name: _Key, min: _Value, max: _Value) -> int: ...
    def zremrangebyrank(self, name: _Key, min: int, max: int) -> int: ...
    def zremrangebyscore(self, name: _Key, min: _Value, max: _Value) -> int: ...
    def zrevrank(self, name: _Key, value: _Value) -> int | None: ...
    def zscore(self, name: _Key, value: _Value) -> float | None: ...
    def zunion(self, keys, aggregate: Any | None = ..., withscores: bool = ...): ...
    def zunionstore(self, dest: _Key, keys: Iterable[_Key], aggregate: Literal["SUM", "MIN", "MAX"] | None = ...) -> int: ...
    def zmscore(self, key, members): ...

class HyperlogCommands:
    def pfadd(self, name: _Key, *values: _Value) -> int: ...
    def pfcount(self, name: _Key) -> int: ...
    def pfmerge(self, dest: _Key, *sources: _Key) -> bool: ...

class HashCommands(Generic[_StrType]):
    def hdel(self, name: _Key, *keys: _Key) -> int: ...
    def hexists(self, name: _Key, key: _Key) -> bool: ...
    def hget(self, name: _Key, key: _Key) -> _StrType | None: ...
    def hgetall(self, name: _Key) -> dict[_StrType, _StrType]: ...
    def hincrby(self, name: _Key, key: _Key, amount: int = ...) -> int: ...
    def hincrbyfloat(self, name: _Key, key: _Key, amount: float = ...) -> float: ...
    def hkeys(self, name: _Key) -> list[_StrType]: ...
    def hlen(self, name: _Key) -> int: ...
    @overload
    def hset(self, name: _Key, key: _Key, value: _Value, mapping: Mapping[_Key, _Value] | None = ...) -> int: ...
    @overload
    def hset(self, name: _Key, key: None, value: None, mapping: Mapping[_Key, _Value]) -> int: ...
    @overload
    def hset(self, name: _Key, *, mapping: Mapping[_Key, _Value]) -> int: ...
    def hsetnx(self, name: _Key, key: _Key, value: _Value) -> int: ...
    def hmset(self, name: _Key, mapping: Mapping[_Key, _Value]) -> bool: ...
    def hmget(self, name: _Key, keys: _Key | Iterable[_Key], *args: _Key) -> list[_StrType | None]: ...
    def hvals(self, name: _Key) -> list[_StrType]: ...
    def hstrlen(self, name, key): ...

class PubSubCommands:
    def publish(self, channel: _Key, message: _Key, **kwargs: _CommandOptions) -> int: ...
    def pubsub_channels(self, pattern: _Key = ..., **kwargs: _CommandOptions) -> list[str]: ...
    def pubsub_numpat(self, **kwargs: _CommandOptions) -> int: ...
    def pubsub_numsub(self, *args: _Key, **kwargs: _CommandOptions) -> list[tuple[str, int]]: ...

class ScriptCommands(Generic[_StrType]):
    def eval(self, script, numkeys, *keys_and_args): ...
    def evalsha(self, sha, numkeys, *keys_and_args): ...
    def script_exists(self, *args): ...
    def script_debug(self, *args): ...
    def script_flush(self, sync_type: Any | None = ...): ...
    def script_kill(self): ...
    def script_load(self, script): ...
    def register_script(self, script: str | _StrType) -> Script: ...

class GeoCommands:
    def geoadd(self, name, values, nx: bool = ..., xx: bool = ..., ch: bool = ...): ...
    def geodist(self, name, place1, place2, unit: Any | None = ...): ...
    def geohash(self, name, *values): ...
    def geopos(self, name, *values): ...
    def georadius(
        self,
        name,
        longitude,
        latitude,
        radius,
        unit: Any | None = ...,
        withdist: bool = ...,
        withcoord: bool = ...,
        withhash: bool = ...,
        count: Any | None = ...,
        sort: Any | None = ...,
        store: Any | None = ...,
        store_dist: Any | None = ...,
        any: bool = ...,
    ): ...
    def georadiusbymember(
        self,
        name,
        member,
        radius,
        unit: Any | None = ...,
        withdist: bool = ...,
        withcoord: bool = ...,
        withhash: bool = ...,
        count: Any | None = ...,
        sort: Any | None = ...,
        store: Any | None = ...,
        store_dist: Any | None = ...,
        any: bool = ...,
    ): ...
    def geosearch(
        self,
        name,
        member: Any | None = ...,
        longitude: Any | None = ...,
        latitude: Any | None = ...,
        unit: str = ...,
        radius: Any | None = ...,
        width: Any | None = ...,
        height: Any | None = ...,
        sort: Any | None = ...,
        count: Any | None = ...,
        any: bool = ...,
        withcoord: bool = ...,
        withdist: bool = ...,
        withhash: bool = ...,
    ): ...
    def geosearchstore(
        self,
        dest,
        name,
        member: Any | None = ...,
        longitude: Any | None = ...,
        latitude: Any | None = ...,
        unit: str = ...,
        radius: Any | None = ...,
        width: Any | None = ...,
        height: Any | None = ...,
        sort: Any | None = ...,
        count: Any | None = ...,
        any: bool = ...,
        storedist: bool = ...,
    ): ...

class ModuleCommands:
    def module_load(self, path, *args): ...
    def module_unload(self, name): ...
    def module_list(self): ...
    def command_info(self): ...
    def command_count(self): ...
    def command_getkeys(self, *args): ...
    def command(self): ...

class Script:
    def __init__(self, registered_client, script) -> None: ...
    def __call__(self, keys=..., args=..., client: Any | None = ...): ...

class BitFieldOperation:
    def __init__(self, client, key, default_overflow: Any | None = ...): ...
    def reset(self) -> None: ...
    def overflow(self, overflow): ...
    def incrby(self, fmt, offset, increment, overflow: Any | None = ...): ...
    def get(self, fmt, offset): ...
    def set(self, fmt, offset, value): ...
    @property
    def command(self): ...
    def execute(self): ...

class ClusterCommands:
    def cluster(self, cluster_arg: str, *args, **kwargs: _CommandOptions): ...
    def readwrite(self, **kwargs: _CommandOptions) -> bool: ...
    def readonly(self, **kwargs: _CommandOptions) -> bool: ...

class DataAccessCommands(
    BasicKeyCommands[_StrType],
    HyperlogCommands,
    HashCommands[_StrType],
    GeoCommands,
    ListCommands[_StrType],
    ScanCommands[_StrType],
    SetCommands[_StrType],
    StreamCommands,
    SortedSetCommands[_StrType],
    Generic[_StrType],
): ...
class CoreCommands(
    ACLCommands[_StrType],
    ClusterCommands,
    DataAccessCommands[_StrType],
    ManagementCommands,
    ModuleCommands,
    PubSubCommands,
    ScriptCommands[_StrType],
    Generic[_StrType],
): ...



from _typeshed import Self
from typing import IO, Pattern

from paramiko.message import Message

OPENSSH_AUTH_MAGIC: bytes

def _unpad_openssh(data: bytes) -> bytes: ...

class PKey:
    public_blob: PublicBlob | None
    BEGIN_TAG: Pattern[str]
    END_TAG: Pattern[str]
    def __init__(self, msg: Message | None = ..., data: str | None = ...) -> None: ...
    def asbytes(self) -> bytes: ...
    def __cmp__(self, other: object) -> int: ...
    def __eq__(self, other: object) -> bool: ...
    def get_name(self) -> str: ...
    def get_bits(self) -> int: ...
    def can_sign(self) -> bool: ...
    def get_fingerprint(self) -> bytes: ...
    def get_base64(self) -> str: ...
    def sign_ssh_data(self, data: bytes) -> Message: ...
    def verify_ssh_sig(self, data: bytes, msg: Message) -> bool: ...
    @classmethod
    def from_private_key_file(cls: type[Self], filename: str, password: str | None = ...) -> Self: ...
    @classmethod
    def from_private_key(cls: type[Self], file_obj: IO[str], password: str | None = ...) -> Self: ...
    def write_private_key_file(self, filename: str, password: str | None = ...) -> None: ...
    def write_private_key(self, file_obj: IO[str], password: str | None = ...) -> None: ...
    def load_certificate(self, value: Message | str) -> None: ...

class PublicBlob:
    key_type: str
    key_blob: str
    comment: str
    def __init__(self, type_: str, blob: bytes, comment: str | None = ...) -> None: ...
    @classmethod
    def from_file(cls: type[Self], filename: str) -> Self: ...
    @classmethod
    def from_string(cls: type[Self], string: str) -> Self: ...
    @classmethod
    def from_message(cls: type[Self], message: Message) -> Self: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...



from typing import Any

from stripe import error as error

class Webhook:
    DEFAULT_TOLERANCE: int
    @staticmethod
    def construct_event(payload, sig_header, secret, tolerance=..., api_key: Any | None = ...): ...

class WebhookSignature:
    EXPECTED_SCHEME: str
    @classmethod
    def verify_header(cls, payload, header, secret, tolerance: Any | None = ...): ...



from typing import Any

from braintree.ach_mandate import AchMandate as AchMandate
from braintree.configuration import Configuration as Configuration
from braintree.resource import Resource as Resource
from braintree.us_bank_account_verification import UsBankAccountVerification as UsBankAccountVerification

class UsBankAccount(Resource):
    @staticmethod
    def find(token): ...
    @staticmethod
    def sale(token, transactionRequest): ...
    @staticmethod
    def signature(): ...
    ach_mandate: Any
    verifications: Any
    def __init__(self, gateway, attributes) -> None: ...



from typing import Any, Generic, TypeVar

from ..sql.operators import ColumnOperators
from ..util import memoized_property
from . import util as orm_util
from .interfaces import MapperProperty, PropComparator

_T = TypeVar("_T")

class DescriptorProperty(MapperProperty):
    doc: Any
    uses_objects: bool
    key: Any
    descriptor: Any
    def instrument_class(self, mapper): ...

class CompositeProperty(DescriptorProperty):
    attrs: Any
    composite_class: Any
    active_history: Any
    deferred: Any
    group: Any
    comparator_factory: Any
    info: Any
    def __init__(self, class_, *attrs, **kwargs) -> None: ...
    def instrument_class(self, mapper) -> None: ...
    def do_init(self) -> None: ...
    @memoized_property
    def props(self): ...
    @property
    def columns(self): ...
    def get_history(self, state, dict_, passive=...): ...

    class CompositeBundle(orm_util.Bundle):
        property: Any
        def __init__(self, property_, expr) -> None: ...
        def create_row_processor(self, query, procs, labels): ...

    class Comparator(PropComparator[_T], Generic[_T]):
        __hash__: Any
        @memoized_property
        def clauses(self): ...
        def __clause_element__(self): ...
        @memoized_property
        def expression(self): ...
        def __eq__(self, other) -> ColumnOperators[_T]: ...  # type: ignore[override]
        def __ne__(self, other) -> ColumnOperators[_T]: ...  # type: ignore[override]

class ConcreteInheritedProperty(DescriptorProperty):
    descriptor: Any
    def __init__(self): ...

class SynonymProperty(DescriptorProperty):
    name: Any
    map_column: Any
    descriptor: Any
    comparator_factory: Any
    doc: Any
    info: Any
    def __init__(
        self,
        name,
        map_column: Any | None = ...,
        descriptor: Any | None = ...,
        comparator_factory: Any | None = ...,
        doc: Any | None = ...,
        info: Any | None = ...,
    ) -> None: ...
    @property
    def uses_objects(self): ...
    def get_history(self, *arg, **kw): ...
    parent: Any
    def set_parent(self, parent, init) -> None: ...



from typing import Any, AnyStr, Callable, Sequence, TypeVar

def with_repr(attrs: Sequence[AnyStr | Attribute]) -> Callable[..., Any]: ...
def with_cmp(attrs: Sequence[AnyStr | Attribute]) -> Callable[..., Any]: ...
def with_init(attrs: Sequence[AnyStr | Attribute]) -> Callable[..., Any]: ...
def immutable(attrs: Sequence[AnyStr | Attribute]) -> Callable[..., Any]: ...
def strip_leading_underscores(attribute_name: AnyStr) -> AnyStr: ...

NOTHING = Any

_T = TypeVar("_T")

def attributes(
    attrs: Sequence[AnyStr | Attribute],
    apply_with_cmp: bool = ...,
    apply_with_init: bool = ...,
    apply_with_repr: bool = ...,
    apply_immutable: bool = ...,
    store_attributes: Callable[[type, Attribute], Any] | None = ...,
    **kw: dict[Any, Any] | None,
) -> Callable[[type[_T]], type[_T]]: ...

class Attribute:
    def __init__(
        self,
        name: AnyStr,
        exclude_from_cmp: bool = ...,
        exclude_from_init: bool = ...,
        exclude_from_repr: bool = ...,
        exclude_from_immutable: bool = ...,
        default_value: Any = ...,
        default_factory: Callable[[None], Any] | None = ...,
        instance_of: Any | None = ...,
        init_aliaser: Callable[[AnyStr], AnyStr] | None = ...,
    ) -> None: ...



import threading
from _typeshed import StrPath, SupportsWrite
from time import struct_time
from types import FrameType, TracebackType
from typing import IO, Any, Callable, Generic, Mapping, MutableMapping, Sequence, Text, TypeVar, Union, overload

_SysExcInfoType = Union[tuple[type, BaseException, TracebackType | None], tuple[None, None, None]]
_ExcInfoType = None | bool | _SysExcInfoType
_ArgsType = Union[tuple[Any, ...], Mapping[str, Any]]
_FilterType = Filter | Callable[[LogRecord], int]
_Level = int | Text

raiseExceptions: bool
logThreads: bool
logMultiprocessing: bool
logProcesses: bool
_srcfile: str | None

def currentframe() -> FrameType: ...

_levelNames: dict[int | str, str | int]  # Union[int:str, str:int]

class Filterer(object):
    filters: list[Filter]
    def __init__(self) -> None: ...
    def addFilter(self, filter: _FilterType) -> None: ...
    def removeFilter(self, filter: _FilterType) -> None: ...
    def filter(self, record: LogRecord) -> bool: ...

class Logger(Filterer):
    name: str
    level: int
    parent: Logger | PlaceHolder
    propagate: bool
    handlers: list[Handler]
    disabled: int
    def __init__(self, name: str, level: _Level = ...) -> None: ...
    def setLevel(self, level: _Level) -> None: ...
    def isEnabledFor(self, level: int) -> bool: ...
    def getEffectiveLevel(self) -> int: ...
    def getChild(self, suffix: str) -> Logger: ...
    def debug(
        self, msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any
    ) -> None: ...
    def info(
        self, msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any
    ) -> None: ...
    def warning(
        self, msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any
    ) -> None: ...
    warn = warning
    def error(
        self, msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any
    ) -> None: ...
    def critical(
        self, msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any
    ) -> None: ...
    fatal = critical
    def log(
        self, level: int, msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any
    ) -> None: ...
    def exception(
        self, msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any
    ) -> None: ...
    def _log(
        self, level: int, msg: Any, args: _ArgsType, exc_info: _ExcInfoType | None = ..., extra: dict[str, Any] | None = ...
    ) -> None: ...  # undocumented
    def filter(self, record: LogRecord) -> bool: ...
    def addHandler(self, hdlr: Handler) -> None: ...
    def removeHandler(self, hdlr: Handler) -> None: ...
    def findCaller(self) -> tuple[str, int, str]: ...
    def handle(self, record: LogRecord) -> None: ...
    def makeRecord(
        self,
        name: str,
        level: int,
        fn: str,
        lno: int,
        msg: Any,
        args: _ArgsType,
        exc_info: _SysExcInfoType | None,
        func: str | None = ...,
        extra: Mapping[str, Any] | None = ...,
    ) -> LogRecord: ...

CRITICAL: int
FATAL: int
ERROR: int
WARNING: int
WARN: int
INFO: int
DEBUG: int
NOTSET: int

class Handler(Filterer):
    level: int  # undocumented
    formatter: Formatter | None  # undocumented
    lock: threading.Lock | None  # undocumented
    name: str | None  # undocumented
    def __init__(self, level: _Level = ...) -> None: ...
    def createLock(self) -> None: ...
    def acquire(self) -> None: ...
    def release(self) -> None: ...
    def setLevel(self, level: _Level) -> None: ...
    def setFormatter(self, fmt: Formatter) -> None: ...
    def filter(self, record: LogRecord) -> bool: ...
    def flush(self) -> None: ...
    def close(self) -> None: ...
    def handle(self, record: LogRecord) -> None: ...
    def handleError(self, record: LogRecord) -> None: ...
    def format(self, record: LogRecord) -> str: ...
    def emit(self, record: LogRecord) -> None: ...

class Formatter:
    converter: Callable[[float | None], struct_time]
    _fmt: str | None
    datefmt: str | None
    def __init__(self, fmt: str | None = ..., datefmt: str | None = ...) -> None: ...
    def format(self, record: LogRecord) -> str: ...
    def formatTime(self, record: LogRecord, datefmt: str | None = ...) -> str: ...
    def formatException(self, ei: _SysExcInfoType) -> str: ...

class Filter:
    def __init__(self, name: str = ...) -> None: ...
    def filter(self, record: LogRecord) -> bool: ...

class LogRecord:
    args: _ArgsType
    asctime: str
    created: int
    exc_info: _SysExcInfoType | None
    exc_text: str | None
    filename: str
    funcName: str
    levelname: str
    levelno: int
    lineno: int
    module: str
    msecs: int
    message: str
    msg: str
    name: str
    pathname: str
    process: int
    processName: str
    relativeCreated: int
    thread: int
    threadName: str
    def __init__(
        self,
        name: str,
        level: int,
        pathname: str,
        lineno: int,
        msg: Any,
        args: _ArgsType,
        exc_info: _SysExcInfoType | None,
        func: str | None = ...,
    ) -> None: ...
    def getMessage(self) -> str: ...

_L = TypeVar("_L", Logger, LoggerAdapter[Logger], LoggerAdapter[Any])

class LoggerAdapter(Generic[_L]):
    logger: _L
    extra: Mapping[str, Any]
    def __init__(self, logger: _L, extra: Mapping[str, Any]) -> None: ...
    def process(self, msg: Any, kwargs: MutableMapping[str, Any]) -> tuple[Any, MutableMapping[str, Any]]: ...
    def debug(
        self, msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any
    ) -> None: ...
    def info(
        self, msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any
    ) -> None: ...
    def warning(
        self, msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any
    ) -> None: ...
    def error(
        self, msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any
    ) -> None: ...
    def exception(
        self, msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any
    ) -> None: ...
    def critical(
        self, msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any
    ) -> None: ...
    def log(
        self, level: int, msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any
    ) -> None: ...
    def isEnabledFor(self, level: int) -> bool: ...

@overload
def getLogger() -> Logger: ...
@overload
def getLogger(name: Text | str) -> Logger: ...
def getLoggerClass() -> type: ...
def debug(msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any) -> None: ...
def info(msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any) -> None: ...
def warning(msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any) -> None: ...

warn = warning

def error(msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any) -> None: ...
def critical(msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any) -> None: ...
def exception(msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any) -> None: ...
def log(
    level: int, msg: Any, *args: Any, exc_info: _ExcInfoType = ..., extra: dict[str, Any] | None = ..., **kwargs: Any
) -> None: ...

fatal = critical

def disable(level: int) -> None: ...
def addLevelName(level: int, levelName: str) -> None: ...
def getLevelName(level: int | str) -> Any: ...
def makeLogRecord(dict: Mapping[str, Any]) -> LogRecord: ...
@overload
def basicConfig() -> None: ...
@overload
def basicConfig(
    *,
    filename: str | None = ...,
    filemode: str = ...,
    format: str = ...,
    datefmt: str | None = ...,
    level: _Level | None = ...,
    stream: IO[str] = ...,
) -> None: ...
def shutdown(handlerList: Sequence[Any] = ...) -> None: ...  # handlerList is undocumented
def setLoggerClass(klass: type) -> None: ...
def captureWarnings(capture: bool) -> None: ...

_StreamT = TypeVar("_StreamT", bound=SupportsWrite[str])

class StreamHandler(Handler, Generic[_StreamT]):
    stream: _StreamT  # undocumented
    @overload
    def __init__(self: StreamHandler[IO[str]], stream: None = ...) -> None: ...
    @overload
    def __init__(self: StreamHandler[_StreamT], stream: _StreamT) -> None: ...

class FileHandler(StreamHandler):
    baseFilename: str  # undocumented
    mode: str  # undocumented
    encoding: str | None  # undocumented
    delay: bool  # undocumented
    def __init__(self, filename: StrPath, mode: str = ..., encoding: str | None = ..., delay: bool = ...) -> None: ...
    def _open(self) -> IO[Any]: ...

class NullHandler(Handler): ...

class PlaceHolder:
    def __init__(self, alogger: Logger) -> None: ...
    def append(self, alogger: Logger) -> None: ...

# Below aren't in module docs but still visible

class RootLogger(Logger):
    def __init__(self, level: int) -> None: ...

root: RootLogger

BASIC_FORMAT: str



from collections.abc import Iterator, Mapping
from typing import Any

from pygments.style import StyleMeta
from pygments.util import ClassNotFound as ClassNotFound

STYLE_MAP: Mapping[str, str]

def get_style_by_name(name) -> StyleMeta: ...
def get_all_styles() -> Iterator[str]: ...

# Having every style class here doesn't seem to be worth it
def __getattr__(name: str) -> Any: ...



from typing import Any, Iterable

__version__: str

_Argv = Iterable[str] | str

def docopt(
    doc: str, argv: _Argv | None = ..., help: bool = ..., version: Any | None = ..., options_first: bool = ...
) -> dict[str, Any]: ...  # Really should be dict[str, str | bool]



from selenium.webdriver.common.desired_capabilities import DesiredCapabilities as DesiredCapabilities
from selenium.webdriver.remote.webdriver import WebDriver as RemoteWebDriver

class WebDriver(RemoteWebDriver):
    def __init__(self, host: str = ..., port: int = ..., desired_capabilities=...) -> None: ...



from _typeshed import StrOrBytesPath
from types import FrameType, TracebackType
from typing import IO, Any, Callable

_ExcInfo = tuple[type[BaseException] | None, BaseException | None, TracebackType | None]

__UNDEF__: object  # undocumented sentinel

def reset() -> str: ...  # undocumented
def small(text: str) -> str: ...  # undocumented
def strong(text: str) -> str: ...  # undocumented
def grey(text: str) -> str: ...  # undocumented
def lookup(name: str, frame: FrameType, locals: dict[str, Any]) -> tuple[str | None, Any]: ...  # undocumented
def scanvars(
    reader: Callable[[], bytes], frame: FrameType, locals: dict[str, Any]
) -> list[tuple[str, str | None, Any]]: ...  # undocumented
def html(einfo: _ExcInfo, context: int = ...) -> str: ...
def text(einfo: _ExcInfo, context: int = ...) -> str: ...

class Hook:  # undocumented
    def __init__(
        self,
        display: int = ...,
        logdir: StrOrBytesPath | None = ...,
        context: int = ...,
        file: IO[str] | None = ...,
        format: str = ...,
    ) -> None: ...
    def __call__(self, etype: type[BaseException] | None, evalue: BaseException | None, etb: TracebackType | None) -> None: ...
    def handle(self, info: _ExcInfo | None = ...) -> None: ...

def handler(info: _ExcInfo | None = ...) -> None: ...
def enable(display: int = ..., logdir: StrOrBytesPath | None = ..., context: int = ..., format: str = ...) -> None: ...



from typing import Pattern, Text
from typing_extensions import Literal

_DEFAULT_DELIMITER: str

def emojize(
    string: str,
    use_aliases: bool = ...,
    delimiters: tuple[str, str] = ...,
    variant: Literal["text_type", "emoji_type", None] = ...,
    language: str = ...,
) -> str: ...
def demojize(string: str, use_aliases: bool = ..., delimiters: tuple[str, str] = ..., language: str = ...) -> str: ...
def get_emoji_regexp(language: str = ...) -> Pattern[Text]: ...
def emoji_lis(string: str, language: str = ...) -> list[dict[str, int | str]]: ...
def distinct_emoji_lis(string: str) -> list[str]: ...
def emoji_count(string: str) -> int: ...



import collections
from _typeshed import Self
from typing import Any, Generic, Iterable, Iterator, Mapping, TypeVar, overload

_KT = TypeVar("_KT")
_VT = TypeVar("_VT")

class frozendict(Mapping[_KT, _VT], Generic[_KT, _VT]):

    dict_cls: type[dict[Any, Any]] = ...
    @overload
    def __init__(self, **kwargs: _VT) -> None: ...
    @overload
    def __init__(self, mapping: Mapping[_KT, _VT]) -> None: ...
    @overload
    def __init__(self, iterable: Iterable[tuple[_KT, _VT]]) -> None: ...
    def __getitem__(self, key: _KT) -> _VT: ...
    def __contains__(self, __key: object) -> bool: ...
    def copy(self: Self, **add_or_replace: _VT) -> Self: ...
    def __iter__(self) -> Iterator[_KT]: ...
    def __len__(self) -> int: ...
    def __hash__(self) -> int: ...

class FrozenOrderedDict(frozendict[_KT, _VT]):

    dict_cls: type[collections.OrderedDict[Any, Any]] = ...



from typing import Any

class Service:
    path: Any
    port: Any
    start_error_message: Any
    log_file: Any
    env: Any
    def __init__(
        self, executable, port: int = ..., log_file=..., env: Any | None = ..., start_error_message: str = ...
    ) -> None: ...
    @property
    def service_url(self): ...
    def command_line_args(self) -> None: ...
    process: Any
    def start(self) -> None: ...
    def assert_process_still_running(self) -> None: ...
    def is_connectable(self): ...
    def send_remote_shutdown_command(self) -> None: ...
    def stop(self) -> None: ...
    def __del__(self) -> None: ...



from typing import Any, Iterable, overload

class LazyFixture:
    name: str
    def __init__(self, name: str) -> None: ...
    def __eq__(self, other: object) -> bool: ...

@overload
def lazy_fixture(names: str) -> LazyFixture: ...
@overload
def lazy_fixture(names: Iterable[str]) -> list[LazyFixture] | Any: ...
def is_lazy_fixture(val: Any) -> bool: ...
def pytest_configure() -> None: ...
def __getattr__(name: str) -> Any: ...  # incomplete



import abc
from typing import Any

bootstrapped_as_sqlalchemy: bool
log: Any
py3k: Any
ABC = abc.ABC

fixtures: Any
engines: Any
exclusions: Any
warnings: Any
profiling: Any
provision: Any
assertions: Any
requirements: Any
config: Any
testing: Any
util: Any
file_config: Any
include_tags: Any
exclude_tags: Any
options: Any

def setup_options(make_option) -> None: ...
def configure_follower(follower_ident) -> None: ...
def memoize_important_follower_config(dict_) -> None: ...
def restore_important_follower_config(dict_) -> None: ...
def read_config() -> None: ...
def pre_begin(opt) -> None: ...
def set_coverage_flag(value) -> None: ...
def post_begin() -> None: ...

pre_configure: Any
post_configure: Any

def pre(fn): ...
def post(fn): ...
def want_class(name, cls): ...
def want_method(cls, fn): ...
def generate_sub_tests(cls, module) -> None: ...
def start_test_class_outside_fixtures(cls) -> None: ...
def stop_test_class(cls) -> None: ...
def stop_test_class_outside_fixtures(cls) -> None: ...
def final_process_cleanup() -> None: ...
def before_test(test, test_module_name, test_class, test_name) -> None: ...
def after_test(test) -> None: ...
def after_test_fixtures(test) -> None: ...

class FixtureFunctions(ABC, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def skip_test_exception(self, *arg, **kw): ...
    @abc.abstractmethod
    def combinations(self, *args, **kw): ...
    @abc.abstractmethod
    def param_ident(self, *args, **kw): ...
    @abc.abstractmethod
    def fixture(self, *arg, **kw): ...
    def get_current_test_name(self) -> None: ...
    @abc.abstractmethod
    def mark_base_test_class(self): ...

def set_fixture_functions(fixture_fn_class) -> None: ...



"""Overlay for third-party chex.dataclass decorator.

See https://github.com/deepmind/chex#dataclass-dataclasspy. Typing-wise, the
differences between @dataclasses.dataclass and @chex.dataclass are:
* The latter has a mappable_dataclass parameter, defaulting to True, which makes
  the dataclass inherit from Mapping.
* Chex dataclasses have replace, from_tuple, and to_tuple methods.
"""

from pytype.abstract import abstract
from pytype.overlays import classgen
from pytype.overlays import dataclass_overlay
from pytype.overlays import overlay
from pytype.overlays import overlay_utils
from pytype.pytd import pytd


class ChexOverlay(overlay.Overlay):

  def __init__(self, ctx):
    member_map = {
        "dataclass": Dataclass.make,
    }
    ast = ctx.loader.import_name("chex")
    super().__init__(ctx, "chex", member_map, ast)


class Dataclass(dataclass_overlay.Dataclass):
  """Implements the @dataclass decorator."""

  _DEFAULT_ARGS = {**dataclass_overlay.Dataclass._DEFAULT_ARGS,
                   "mappable_dataclass": True}

  @classmethod
  def make(cls, ctx):
    return super().make(ctx, "chex")

  def _add_replace_method(self, node, cls):
    cls.members["replace"] = classgen.make_replace_method(
        self.ctx, node, cls, kwargs_name="changes")

  def _add_from_tuple_method(self, node, cls):
    # from_tuple is discouraged anyway, so we provide only bare-bones types.
    cls.members["from_tuple"] = overlay_utils.make_method(
        ctx=self.ctx,
        node=node,
        name="from_tuple",
        params=[overlay_utils.Param("args")],
        return_type=cls,
        kind=pytd.MethodKind.STATICMETHOD,
    )

  def _add_to_tuple_method(self, node, cls):
    # to_tuple is discouraged anyway, so we provide only bare-bones types.
    cls.members["to_tuple"] = overlay_utils.make_method(
        ctx=self.ctx,
        node=node,
        name="to_tuple",
        return_type=self.ctx.convert.tuple_type,
    )

  def decorate(self, node, cls):
    super().decorate(node, cls)
    if not isinstance(cls, abstract.InterpreterClass):
      return
    self._add_replace_method(node, cls)
    self._add_from_tuple_method(node, cls)
    self._add_to_tuple_method(node, cls)
    if not self.args[cls]["mappable_dataclass"]:
      return
    mapping = self.ctx.convert.name_to_value("typing.Mapping")
    overlay_utils.add_base_class(node, cls, mapping)



from typing import Any

# Enable when pyasn1 gets stubs:
# from pyasn1.type.univ import Integer, OctetString, Sequence
Integer = Any
OctetString = Any
Sequence = Any

MAXINT: Any
rangeInt0ToMaxConstraint: Any

class Integer0ToMax(Integer):
    subtypeSpec: Any

class Size(Integer0ToMax): ...
class Cookie(OctetString): ...

class RealSearchControlValue(Sequence):
    componentType: Any

def paged_search_control(criticality: bool = ..., size: int = ..., cookie: Any | None = ...): ...



from typing import Any

from selenium.webdriver.common import service as service

class Service(service.Service):
    service_args: Any
    def __init__(
        self, executable_path, port: int = ..., service_args: Any | None = ..., log_path: str = ..., env: Any | None = ...
    ) -> None: ...
    def command_line_args(self): ...
    def send_remote_shutdown_command(self) -> None: ...



import os
import pickle

from pytype import config
from pytype import file_utils
from pytype import load_pytd
from pytype.pytd import pytd_utils
from pytype.pytd import serialize_ast
from pytype.pytd import visitors
from pytype.tests import test_base

import unittest


class SerializeAstTest(test_base.UnitTest):

  def _store_ast(
      self, temp_dir, module_name, pickled_ast_filename, ast=None, loader=None,
      is_package=False, metadata=None):
    if not (ast and loader):
      ast, loader = self._get_ast(temp_dir=temp_dir, module_name=module_name)
    serialize_ast.StoreAst(
        ast,
        pickled_ast_filename,
        is_package=is_package,
        metadata=metadata)
    module_map = {name: module.ast
                  for name, module in loader._modules.items()}

    return module_map

  def _get_ast(self, temp_dir, module_name, src=None):
    src = src or ("""
        import module2
        from module2 import f
        from typing import List

        constant = True

        x = List[int]
        b = List[int]

        class SomeClass:
          def __init__(self, a: module2.ObjectMod2):
            pass

        def ModuleFunction():
          pass
    """)
    pyi_filename = temp_dir.create_file("module1.pyi", src)
    temp_dir.create_file("module2.pyi", """
        import queue
        def f() -> queue.Queue: ...
        class ObjectMod2:
          def __init__(self):
            pass
    """)

    loader = load_pytd.Loader(config.Options.create(
        python_version=self.python_version, pythonpath=temp_dir.path))
    ast = loader.load_file(module_name, pyi_filename)
    # serialize_ast.StoreAst sorts the ast for determinism, so we should do the
    # same to the original ast to do pre- and post-pickling comparisons.
    loader._modules[module_name].ast = ast = ast.Visit(
        visitors.CanonicalOrderingVisitor())
    return ast, loader

  def test_find_class_types_visitor(self):
    module_name = "foo.bar"
    with file_utils.Tempdir() as d:
      ast, _ = self._get_ast(temp_dir=d, module_name=module_name)
    indexer = serialize_ast.FindClassTypesVisitor()
    ast.Visit(indexer)

    self.assertEqual(len(indexer.class_type_nodes), 10)

  def test_node_index_visitor_usage(self):
    """Confirms that the node index is used.

    This removes the first node from the class_type_nodes list and checks that
    that node is not updated by ProcessAst.
    """
    with file_utils.Tempdir() as d:
      module_name = "module1"
      pickled_ast_filename = os.path.join(d.path, "module1.pyi.pickled")
      module_map = self._store_ast(d, module_name, pickled_ast_filename)
      del module_map[module_name]
      serialized_ast = pytd_utils.LoadPickle(pickled_ast_filename)

      # The sorted makes the testcase more deterministic.
      serialized_ast = serialized_ast.Replace(class_type_nodes=sorted(
          serialized_ast.class_type_nodes)[1:])
      loaded_ast = serialize_ast.ProcessAst(serialized_ast, module_map)

      with self.assertRaisesRegex(
          ValueError, "Unresolved class: 'builtins.NoneType'"):
        loaded_ast.Visit(visitors.VerifyLookup())

  def test_pickle(self):
    with file_utils.Tempdir() as d:
      ast, _ = self._get_ast(temp_dir=d, module_name="foo.bar.module1")
      pickled_ast_filename = os.path.join(d.path, "module1.pyi.pickled")

      result = serialize_ast.StoreAst(ast, pickled_ast_filename)

      self.assertIsNone(result)
      with open(pickled_ast_filename, "rb") as fi:
        serialized_ast = pickle.load(fi)
      self.assertTrue(serialized_ast.ast)
      self.assertCountEqual(
          dict(serialized_ast.dependencies),
          ["builtins", "foo.bar.module1", "module2", "queue"])

  def test_unrestorable_child(self):
    # Assume .cls in a ClassType X in module1 was referencing something for
    # which, Visitors.LookupExternalTypes returned AnythingType.
    # Now if something in module1 is referencing X external types need to be
    # resolved before local types, so that we can resolve local types to the
    # correct ClassType, as the ClassType instance changes, if .cls can not be
    # filled and instead AnythingType is used.

    class RenameVisitor(visitors.Visitor):

      def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._init = False

      def EnterFunction(self, func):
        if func.name == "__init__":
          self._init = True
          return None
        return False

      def LeaveFunction(self, func):
        self._init = False

      def VisitClassType(self, cls_type):
        if self._init:
          cls_type = cls_type.Replace(name="other_module.unknown_Reference")
          # Needs to be copied manually as it is not part of the NamedTuple.
          cls_type.cls = None
        return cls_type

    with file_utils.Tempdir() as d:
      src = ("""
        import other_module
        x = other_module.UnusedReferenceNeededToKeepTheImport

        class SomeClass:
          def __init__(will_be_replaced_with_visitor) -> None:
            pass

        def func(a:SomeClass) -> None:
          pass
      """)
      d.create_file("other_module.pyi", """
          from typing import Any
          def __getattr__(self, name) -> Any: ..."""
                   )
      ast, loader = self._get_ast(
          temp_dir=d, module_name="module1", src=src)

      ast = ast.Visit(RenameVisitor())

      pickled_ast_filename = os.path.join(d.path, "module1.pyi.pickled")
      module_map = self._store_ast(
          d, "module1", pickled_ast_filename, ast=ast, loader=loader)
      del module_map["module1"]

      serialized_ast = pytd_utils.LoadPickle(pickled_ast_filename)
      loaded_ast = serialize_ast.ProcessAst(
          serialized_ast, module_map)
      # Look up the "SomeClass" in "def func(a: SomeClass), then run
      # VerifyLookup on it. We can't run VerifyLookup on the root node, since
      # visitors don't descend into the "cls" attribute of ClassType.
      cls = loaded_ast.functions[0].signatures[0].params[0].type.cls
      cls.Visit(visitors.VerifyLookup())

  def test_load_top_level(self):
    """Tests that a pickled file can be read."""
    with file_utils.Tempdir() as d:
      module_name = "module1"
      pickled_ast_filename = os.path.join(d.path, "module1.pyi.pickled")
      module_map = self._store_ast(d, module_name, pickled_ast_filename)
      original_ast = module_map[module_name]
      del module_map[module_name]
      loaded_ast = serialize_ast.ProcessAst(
          pytd_utils.LoadPickle(pickled_ast_filename),
          module_map)

      self.assertTrue(loaded_ast)
      self.assertIsNot(loaded_ast, original_ast)
      self.assertEqual(loaded_ast.name, module_name)
      self.assertTrue(pytd_utils.ASTeq(original_ast, loaded_ast))
      loaded_ast.Visit(visitors.VerifyLookup())

  def test_load_with_same_module_name(self):
    """Explicitly set the module name and reload with the same name.

    The difference to testLoadTopLevel is that the module name does not match
    the filelocation.
    """
    with file_utils.Tempdir() as d:
      module_name = "foo.bar.module1"
      pickled_ast_filename = os.path.join(d.path, "module1.pyi.pickled")
      module_map = self._store_ast(d, module_name, pickled_ast_filename)
      original_ast = module_map[module_name]
      del module_map[module_name]

      loaded_ast = serialize_ast.ProcessAst(
          pytd_utils.LoadPickle(pickled_ast_filename),
          module_map)

      self.assertTrue(loaded_ast)
      self.assertIsNot(loaded_ast, original_ast)
      self.assertEqual(loaded_ast.name, "foo.bar.module1")
      self.assertTrue(pytd_utils.ASTeq(original_ast, loaded_ast))
      loaded_ast.Visit(visitors.VerifyLookup())

  def test_unrestorable_dependency_error_with_module_index(self):
    with file_utils.Tempdir() as d:
      module_name = "module1"
      pickled_ast_filename = os.path.join(d.path, "module1.pyi.pickled")
      module_map = self._store_ast(d, module_name, pickled_ast_filename)
      module_map = {}  # Remove module2

      with self.assertRaises(serialize_ast.UnrestorableDependencyError):
        serialize_ast.ProcessAst(
            pytd_utils.LoadPickle(pickled_ast_filename),
            module_map)

  def test_unrestorable_dependency_error_without_module_index(self):
    with file_utils.Tempdir() as d:
      module_name = "module1"
      pickled_ast_filename = os.path.join(d.path, "module1.pyi.pickled")
      module_map = self._store_ast(d, module_name, pickled_ast_filename)
      module_map = {}  # Remove module2

      loaded_ast = pytd_utils.LoadPickle(pickled_ast_filename)
      loaded_ast.modified_class_types = None  # Remove the index
      with self.assertRaises(serialize_ast.UnrestorableDependencyError):
        serialize_ast.ProcessAst(loaded_ast, module_map)

  def test_load_with_different_module_name(self):
    with file_utils.Tempdir() as d:
      original_module_name = "module1"
      pickled_ast_filename = os.path.join(d.path, "module1.pyi.pickled")
      module_map = self._store_ast(
          d, original_module_name, pickled_ast_filename)
      original_ast = module_map[original_module_name]
      del module_map[original_module_name]

      new_module_name = "wurstbrot.module2"
      serializable_ast = pytd_utils.LoadPickle(pickled_ast_filename)
      serializable_ast = serialize_ast.EnsureAstName(
          serializable_ast, new_module_name, fix=True)
      loaded_ast = serialize_ast.ProcessAst(serializable_ast, module_map)

      self.assertTrue(loaded_ast)
      self.assertIsNot(loaded_ast, original_ast)
      self.assertEqual(loaded_ast.name, new_module_name)
      loaded_ast.Visit(visitors.VerifyLookup())
      self.assertFalse(pytd_utils.ASTeq(original_ast, loaded_ast))
      ast_new_module, _ = self._get_ast(temp_dir=d, module_name=new_module_name)
      self.assertTrue(pytd_utils.ASTeq(ast_new_module, loaded_ast))

  def test_store_removes_init(self):
    with file_utils.Tempdir() as d:
      original_module_name = "module1.__init__"
      pickled_ast_filename = os.path.join(d.path, "module1.pyi.pickled")

      module_map = self._store_ast(
          d, original_module_name, pickled_ast_filename, is_package=True)
      serializable_ast = pytd_utils.LoadPickle(pickled_ast_filename)

      expected_name = "module1"
      # Check that the module had the expected name before.
      self.assertIn(original_module_name, module_map)
      # Check that module1 wasn't created before storing.
      self.assertNotIn(expected_name, module_map)
      # Check that the saved ast had its name changed.
      self.assertEqual(serializable_ast.ast.name, expected_name)

  def test_function(self):
    with file_utils.Tempdir() as d:
      foo = d.create_file("foo.pickle")
      module_map = self._store_ast(d, "foo", foo, ast=self._get_ast(d, "foo"))
      p = pytd_utils.LoadPickle(foo)
      ast = serialize_ast.ProcessAst(p, module_map)
      f, = (a for a in ast.aliases if a.name == "foo.f")
      signature, = f.type.signatures
      self.assertIsNotNone(signature.return_type.cls)

  def test_pickle_metadata(self):
    with file_utils.Tempdir() as d:
      module_name = "module1"
      pickled_ast_filename = os.path.join(d.path, "module1.pyi.pickled")
      self._store_ast(
          d, module_name, pickled_ast_filename, metadata=["meta", "data"])
      serialized_ast = pytd_utils.LoadPickle(pickled_ast_filename)
      self.assertSequenceEqual(serialized_ast.metadata, ["meta", "data"])

if __name__ == "__main__":
  unittest.main()



"""Tests for mixin.py."""

from pytype.abstract import mixin  # pylint: disable=unused-import
import unittest


class MixinMetaTest(unittest.TestCase):

  def test_mixin_super(self):
    """Test the imitation 'super' method on MixinMeta."""
    # pylint: disable=g-wrong-blank-lines,undefined-variable
    class A:
      def f(self, x):
        return x

    class MyMixin(metaclass=mixin.MixinMeta):
      overloads = ("f",)
      def f(self, x):
        if x == 0:
          return "hello"
        return MyMixin.super(self.f)(x)

    class B(A, MyMixin):
      pass
    # pylint: enable=g-wrong-blank-lines,undefined-variable

    b = B()
    v_mixin = b.f(0)
    v_a = b.f(1)
    self.assertEqual(v_mixin, "hello")
    self.assertEqual(v_a, 1)


class PythonDictTest(unittest.TestCase):

  def test_wraps_dict(self):
    # pylint: disable=g-wrong-blank-lines,undefined-variable
    class A(mixin.PythonDict):
      def __init__(self, pyval):
        mixin.PythonDict.init_mixin(self, pyval)
    # pylint: enable=g-wrong-blank-lines,undefined-variable

    a = A({"foo": 1, "bar": 2})
    self.assertEqual(a.get("x", "baz"), "baz")
    self.assertNotIn("x", a)
    self.assertEqual(a.get("foo"), 1)
    self.assertEqual(a["foo"], 1)
    self.assertIn("foo", a)
    self.assertIn("bar", a)
    self.assertEqual(a.copy(), a.pyval)
    self.assertCountEqual(iter(a), ["foo", "bar"])
    self.assertCountEqual(a.keys(), ["foo", "bar"])
    self.assertCountEqual(a.values(), [1, 2])
    self.assertCountEqual(a.items(), [("foo", 1), ("bar", 2)])

if __name__ == "__main__":
  unittest.main()



from typing import Any

def paged_search_generator(
    connection,
    search_base,
    search_filter,
    search_scope=...,
    dereference_aliases=...,
    attributes: Any | None = ...,
    size_limit: int = ...,
    time_limit: int = ...,
    types_only: bool = ...,
    get_operational_attributes: bool = ...,
    controls: Any | None = ...,
    paged_size: int = ...,
    paged_criticality: bool = ...,
) -> None: ...
def paged_search_accumulator(
    connection,
    search_base,
    search_filter,
    search_scope=...,
    dereference_aliases=...,
    attributes: Any | None = ...,
    size_limit: int = ...,
    time_limit: int = ...,
    types_only: bool = ...,
    get_operational_attributes: bool = ...,
    controls: Any | None = ...,
    paged_size: int = ...,
    paged_criticality: bool = ...,
): ...



"""Utilities for checking function overrides used in vm.py."""

import dataclasses
import enum
import logging

from pytype.abstract import abstract
from pytype.abstract import abstract_utils
from pytype.abstract import function
from pytype.pytd import pytd

log = logging.getLogger(__name__)


@enum.unique
class SignatureErrorType(enum.Enum):
  """Constants representing various signature mismatch errors."""
  NO_ERROR = enum.auto()
  DEFAULT_PARAMETER_MISMATCH = enum.auto()
  DEFAULT_VALUE_MISMATCH = enum.auto()
  KWONLY_PARAMETER_COUNT_MISMATCH = enum.auto()
  KWONLY_PARAMETER_NAME_MISMATCH = enum.auto()
  KWONLY_PARAMETER_TYPE_MISMATCH = enum.auto()
  POSITIONAL_PARAMETER_COUNT_MISMATCH = enum.auto()
  POSITIONAL_PARAMETER_NAME_MISMATCH = enum.auto()
  POSITIONAL_PARAMETER_TYPE_MISMATCH = enum.auto()
  RETURN_TYPE_MISMATCH = enum.auto()


SIGNATURE_ERROR_TYPE_TO_OPTION_NAME = {
    SignatureErrorType.DEFAULT_PARAMETER_MISMATCH:
        "overriding_parameter_count_checks",
    SignatureErrorType.DEFAULT_VALUE_MISMATCH:
        "overriding_default_value_checks",
    SignatureErrorType.KWONLY_PARAMETER_COUNT_MISMATCH:
        "overriding_parameter_count_checks",
    SignatureErrorType.KWONLY_PARAMETER_NAME_MISMATCH:
        "overriding_parameter_name_checks",
    SignatureErrorType.KWONLY_PARAMETER_TYPE_MISMATCH:
        "overriding_parameter_type_checks",
    SignatureErrorType.POSITIONAL_PARAMETER_COUNT_MISMATCH:
        "overriding_parameter_count_checks",
    SignatureErrorType.POSITIONAL_PARAMETER_NAME_MISMATCH:
        "overriding_parameter_name_checks",
    SignatureErrorType.POSITIONAL_PARAMETER_TYPE_MISMATCH:
        "overriding_parameter_type_checks",
    SignatureErrorType.RETURN_TYPE_MISMATCH:
        "overriding_return_type_checks",
}


@dataclasses.dataclass
class SignatureError:
  error_code: SignatureErrorType = SignatureErrorType.NO_ERROR
  message: str = ""


def _get_varargs_annotation_type(param_type):
  """Returns the type annotation for the varargs parameter."""
  # If varargs are annotated with type T in the function definition,
  # the annotation in the signature will be Tuple[T, ...].
  if not isinstance(param_type, abstract.ParameterizedClass):
    # Annotation in the signature could be just 'tuple' that corresponds to
    # varargs with no annotation.
    return None
  return param_type.get_formal_type_parameter(abstract_utils.T)


def _check_positional_parameter_annotations(method_signature, base_signature,
                                            is_subtype):
  """Checks type annotations for positional parameters of the overriding method.

  Args:
    method_signature: signature of the overriding method.
    base_signature: signature of the overridden method.
    is_subtype: a binary function to compare types.

  Returns:
    SignatureError if a mismatch is detected. Otherwise returns None.
  """
  for param_index in range(max(len(base_signature.param_names),
                               len(method_signature.param_names))):
    if param_index == 0:
      # No type checks for 'self' parameter.
      continue
    if param_index < len(base_signature.param_names):
      base_param_name = base_signature.param_names[param_index]
    elif base_signature.varargs_name:
      base_param_name = base_signature.varargs_name
    else:
      # Remaining positional parameters aren't mapped. Not checking annotations.
      break

    try:
      base_param_type = base_signature.annotations[base_param_name]
    except KeyError:
      # Parameter not annotated in overridden method.
      continue

    if base_param_name == base_signature.varargs_name:
      base_param_type = _get_varargs_annotation_type(base_param_type)
      if base_param_type is None:
        continue

    if param_index < method_signature.posonly_count:
      # Positional-only parameters should match by position.
      method_param_name = method_signature.param_names[param_index]
    elif param_index < len(method_signature.param_names):
      if (base_param_name == "_" or
          method_signature.param_names[param_index] == "_"):
        # Underscore parameters should match by position.
        method_param_name = method_signature.param_names[param_index]
      else:
        # Other positional-or-keyword parameters should match by name.
        method_param_name = base_param_name
    elif method_signature.varargs_name:
      method_param_name = method_signature.varargs_name
    else:
      # Remaining positional parameters aren't mapped. Not checking annotations.
      break

    try:
      method_param_type = method_signature.annotations[method_param_name]
    except KeyError:
      # Parameter not annotated in overriding method.
      continue

    if method_param_name == method_signature.varargs_name:
      method_param_type = _get_varargs_annotation_type(method_param_type)
      if method_param_type is None:
        continue

    # Parameter type of the overridden method must be a subtype of the
    # parameter type of the overriding method.
    if not is_subtype(base_param_type, method_param_type):
      return SignatureError(
          SignatureErrorType.POSITIONAL_PARAMETER_TYPE_MISMATCH,
          f"Type mismatch for parameter '{method_param_name}'.")

  return None


def _check_positional_parameters(method_signature, base_signature, is_subtype):
  """Checks that the positional parameters of the overriding method match.

  Args:
    method_signature: signature of the overriding method.
    base_signature: signature of the overridden method.
    is_subtype: a binary function to compare types.

  Returns:
    SignatureError if a mismatch is detected. Otherwise returns None.
  """

  # Check mappings of positional-or-keyword parameters of the overridden method.
  for base_param_pos, base_param_name in enumerate(base_signature.param_names):
    # Skip positional-only parameters - those that cannot be passed by name.
    if base_param_pos == 0 or base_param_pos < base_signature.posonly_count:
      continue
    if base_param_name == "_":
      continue

    # Positional-or-keyword cannot map to positional-only.
    if base_param_pos < method_signature.posonly_count:
      return SignatureError(
          SignatureErrorType.POSITIONAL_PARAMETER_COUNT_MISMATCH,
          "Too many positional-only parameters in overriding method.")
    elif base_param_pos < len(method_signature.param_names):
      method_param_name = method_signature.param_names[base_param_pos]
    else:
      # Positional-or-keyword cannot map to keyword-only.
      if method_signature.varargs_name:
        # Remaining parameters map to the varargs parameters.
        break
      return SignatureError(
          SignatureErrorType.POSITIONAL_PARAMETER_COUNT_MISMATCH,
          "Not enough positional parameters in overriding method.")

    # Positional-or-keyword parameters must have the same name or underscore.
    method_param_name = method_signature.param_names[base_param_pos]
    if method_param_name not in (base_param_name, "_"):
      # We don't report it as an error, as this is a very common practice
      # in the absence of positional-only parameters.
      # TODO(sinopalnikov): clean it up and start flagging the error.
      log.warning("Name mismatch for parameter '%r'.", base_param_name)
      return None

  # Check mappings of remaining positional parameters of the overridding method
  # that don't map to any positional parameters of the overridden method.
  remaining_method_params = (
      method_signature.param_names[len(base_signature.param_names):]
      if not base_signature.varargs_name else [])
  for method_param_name in remaining_method_params:

    # Keyword-only can map to remaining positional.
    if method_param_name in base_signature.kwonly_params:
      continue

    # Otherwise remaining positional must have a default value.
    if method_param_name not in method_signature.defaults:
      return SignatureError(
          SignatureErrorType.DEFAULT_PARAMETER_MISMATCH,
          f"Parameter '{method_param_name}' must have a default value.")

  return _check_positional_parameter_annotations(method_signature,
                                                 base_signature, is_subtype)


def _check_keyword_only_parameters(method_signature, base_signature,
                                   is_subtype):
  """Checks that the keyword-only parameters of the overriding method match.

  Args:
    method_signature: signature of the overriding method.
    base_signature: signature of the overridden method.
    is_subtype: a binary function to compare types.

  Returns:
    SignatureError if a mismatch is detected. Otherwise returns None.
  """

  base_kwonly_params = set(base_signature.kwonly_params)
  method_kwonly_params = set(method_signature.kwonly_params)
  method_defaults = set(method_signature.defaults)

  if not base_signature.kwargs_name:
    # Keyword-only parameters of the overriding method that don't match any
    # keyword-only parameter of the overridden method must have a default value.
    for method_param_name in method_kwonly_params.difference(
        base_kwonly_params).difference(method_defaults):
      return SignatureError(
          SignatureErrorType.DEFAULT_PARAMETER_MISMATCH,
          f"Parameter '{method_param_name}' must have a default value.")

  # A keyword-only parameter of the overridden method cannot have the same name
  # as a positional-only parameter of the overriding method.
  for base_param_name in base_kwonly_params.difference(method_kwonly_params):
    try:
      method_param_index = method_signature.param_names.index(base_param_name)
    except ValueError:
      if not method_signature.kwargs_name:
        return SignatureError(
            SignatureErrorType.KWONLY_PARAMETER_NAME_MISMATCH,
            f"Parameter '{base_param_name}' not found in overriding method.")
    else:
      if method_param_index < method_signature.posonly_count:
        return SignatureError(
            SignatureErrorType.KWONLY_PARAMETER_NAME_MISMATCH,
            (f"Keyword-only parameter '{base_param_name}' of the overridden "
             "method has the same name as a positional-only parameter"
             "of the overriding method."))

  # Check annotations of keyword-only parameters.
  for base_param_name in base_signature.kwonly_params:
    try:
      base_param_type = base_signature.annotations[base_param_name]
    except KeyError:
      # Parameter not annotated in the overridden method.
      continue

    if (base_param_name in method_kwonly_params or
        base_param_name in method_signature.param_names):
      method_param_name = base_param_name
    elif method_signature.kwargs_name:
      method_param_name = method_signature.kwargs_name
    else:
      continue

    try:
      method_param_type = method_signature.annotations[method_param_name]
    except KeyError:
      # Parameter not annotated in the overriding method.
      continue

    if method_param_name == method_signature.kwargs_name:
      if isinstance(method_param_type, abstract.ParameterizedClass):
        # If kwargs are annotated with type T in the function definition,
        # the annotation in the signature will be Dict[str, T].
        method_param_type = method_param_type.get_formal_type_parameter(
            abstract_utils.V)
      else:
        # If the kwargs type is a plain dict, there's nothing to check.
        continue

    # Parameter type of the overridden method must be a subtype of the
    # parameter type of the overriding method.
    if not is_subtype(base_param_type, method_param_type):
      return SignatureError(
          SignatureErrorType.KWONLY_PARAMETER_TYPE_MISMATCH,
          f"Type mismatch for parameter '{base_param_name}'.")

  return None


def _check_default_values(method_signature, base_signature):
  """Checks that default parameter values of the overriding method match.

  Args:
    method_signature: signature of the overriding method.
    base_signature: signature of the overridden method.

  Returns:
    SignatureError if a mismatch is detected. Otherwise returns None.
  """

  for base_param_name, base_default_value in base_signature.defaults.items():
    if base_param_name in base_signature.kwonly_params:
      # The parameter is keyword-only, check match by name.
      if (base_param_name not in method_signature.kwonly_params and
          base_param_name not in method_signature.param_names):
        continue
      method_param_name = base_param_name
    else:
      # The parameter is positional, check match by position.
      base_param_index = base_signature.param_names.index(base_param_name)
      if base_param_index >= len(method_signature.param_names):
        continue
      method_param_name = method_signature.param_names[base_param_index]

    try:
      method_default_value = method_signature.defaults[method_param_name]
    except KeyError:
      return SignatureError(
          SignatureErrorType.DEFAULT_PARAMETER_MISMATCH,
          f"Parameter '{method_param_name}' must have a default value.")

    # Only concrete values can be compared for an exact match.
    try:
      base_default = abstract_utils.get_atomic_python_constant(
          base_default_value)
      method_default = abstract_utils.get_atomic_python_constant(
          method_default_value)
    except abstract_utils.ConversionError:
      continue

    if base_default != method_default:
      return SignatureError(
          SignatureErrorType.DEFAULT_VALUE_MISMATCH,
          (f"Parameter '{base_param_name}' must have the same default "
           "value."))

  return None


def _check_return_types(method_signature, base_signature, is_subtype):
  """Checks that the return types match."""
  try:
    base_return_type = base_signature.annotations["return"]
    method_return_type = method_signature.annotations["return"]
  except KeyError:
    # Return type not annotated in either of the two methods.
    return None

  if (isinstance(base_return_type, abstract.AMBIGUOUS_OR_EMPTY) or
      isinstance(method_return_type, abstract.AMBIGUOUS_OR_EMPTY)):
    return None

  # Return type of the overriding method must be a subtype of the
  # return type of the overridden method.
  if not is_subtype(method_return_type, base_return_type):
    return SignatureError(SignatureErrorType.RETURN_TYPE_MISMATCH,
                          "Return type mismatch.")

  return None


def _check_signature_compatible(method_signature, base_signature,
                                stack, matcher, ctx):
  """Checks if the signatures match for the overridden and overriding methods.

  Adds the first error found to the context's error log.

  Two invariants are verified:
  1. Every call that is valid for the overridden method is valid for
     the overriding method.
  2. Two calls that are equivalent for the overridden method are equivalent
     for the overriding method.

  This translates into the following mapping requirements for
  overriding method parameters:
  +----------------------------------------------------------------------------+
  | Overridden method                    | Overriding method                   |
  +----------------------------------------------------------------------------+
  | Positional-only                      | Positional-only or                  |
  |                                      | positional-or-keyword, any name     |
  +--------------------------------------+-------------------------------------+
  | Positional-or-keyword                | Positional-or-keyword, same name    |
  +--------------------------------------+-------------------------------------+
  | Keyword-only                         | Positional-or-keyword               |
  |                                      | or keyword-only, same name          |
  +--------------------------------------+-------------------------------------+
  | Non-default                          | Non-default or default              |
  +--------------------------------------+-------------------------------------+
  | Default                              | Default, same default value         |
  +--------------------------------------+-------------------------------------+
  | Parameter of type T                  | Parameter of supertype of T or      |
  |                                      | no annotation                       |
  +--------------------------------------+-------------------------------------+
  | Parameter without annotation         | Parameter with any type annotation  |
  |                                      | or without annotation               |
  +--------------------------------------+-------------------------------------+
  | Return type T                        | Return type - subtype of T or       |
  |                                      | no annotation                       |
  +--------------------------------------+-------------------------------------+
  | Return type not annotated            | Any return type annotation          |
  |                                      | or no annotation                    |
  +--------------------------------------+-------------------------------------+
  In addition, default parameters of the overriding method don't have to match
  any parameters of the overridden method.
  Same name requirement is often violated, so we don't treat is as an error
  for now and only log a warning.

  Arguments:
    method_signature: signature of the overriding method.
    base_signature: signature of the overridden method.
    stack: the stack to use for mismatch error reporting.
    matcher: abstract matcher for type comparison.
    ctx: pytype abstract context.
  """

  def is_subtype(this_type, that_type):
    """Return True iff this_type is a subclass of that_type."""
    this_type_instance = this_type.instantiate(ctx.root_node, None)
    return not matcher.bad_matches(this_type_instance, that_type)[0]

  check_result = (
      _check_positional_parameters(method_signature, base_signature, is_subtype)
      or _check_keyword_only_parameters(method_signature, base_signature,
                                        is_subtype) or
      _check_default_values(method_signature, base_signature) or
      _check_return_types(method_signature, base_signature, is_subtype))

  if check_result:
    option_name = SIGNATURE_ERROR_TYPE_TO_OPTION_NAME[check_result.error_code]
    if getattr(ctx.options, option_name):
      ctx.errorlog.overriding_signature_mismatch(
          stack, base_signature, method_signature, details=check_result.message)
    else:
      log.warning(check_result.message)


def _get_pytd_class_signature_map(cls, ctx):
  """Returns a map from method names to their signatures for a PyTDClass."""
  assert isinstance(cls, abstract.PyTDClass)
  if cls in ctx.method_signature_map:
    return ctx.method_signature_map[cls]

  method_signature_map = {}
  pytd_cls = cls.pytd_cls
  for pytd_func in pytd_cls.methods:
    if pytd_func.kind != pytd.MethodKind.METHOD:
      continue
    func_name = pytd_func.name
    if func_name in ("__new__", "__init__"):
      continue
    # Assume every method has at least one signature.
    # Ignore overloaded methods, take only the first signature.
    pytd_signature = pytd_func.signatures[0]
    signature = function.Signature.from_pytd(ctx, func_name, pytd_signature)
    assert func_name not in method_signature_map
    method_signature_map[func_name] = signature

  ctx.method_signature_map[cls] = method_signature_map
  return method_signature_map


def check_overriding_members(cls, bases, members, matcher, ctx):
  """Check that the method signatures of the new class match base classes."""

  # Maps method names to methods.
  class_method_map = {}
  for member_name, member_value in members.items():
    try:
      atomic_value = abstract_utils.get_atomic_value(
          member_value, constant_type=abstract.InterpreterFunction)
    except abstract_utils.ConversionError:
      continue
    method = atomic_value
    if method.is_classmethod:
      continue
    method_name = method.name.rsplit(".")[-1]
    if method_name in ("__new__", "__init__"):
      continue
    assert member_name not in class_method_map
    class_method_map[member_name] = method

  class_signature_map = {
      method_name: method.signature
      for method_name, method in class_method_map.items()
  }
  for base in bases:
    try:
      base_class = abstract_utils.get_atomic_value(base)
    except abstract_utils.ConversionError:
      continue
    if isinstance(base_class, abstract.ParameterizedClass):
      base_class = base_class.base_cls
    if isinstance(base_class, abstract.InterpreterClass):
      base_signature_map = ctx.method_signature_map[base_class]
    elif isinstance(base_class, abstract.PyTDClass):
      base_signature_map = _get_pytd_class_signature_map(base_class, ctx)
    else:
      continue

    for base_method_name, base_method_signature in base_signature_map.items():
      if base_method_name not in class_signature_map:
        continue
      class_method_signature = class_signature_map[base_method_name]
      # If the method is defined in the current class then report mismatch
      # errors on the method definition, otherwise report them on the current
      # class definition.
      method_def_opcode = (
          class_method_map[base_method_name].def_opcode
          if base_method_name in class_method_map else None)
      stack = ctx.vm.simple_stack(method_def_opcode)
      _check_signature_compatible(class_method_signature, base_method_signature,
                                  stack, matcher, ctx)

    # Methods defined in this class take precedence.
    class_signature_map = {**base_signature_map, **class_signature_map}

  assert cls not in ctx.method_signature_map
  ctx.method_signature_map[cls] = class_signature_map



import sys
from typing import Callable, Sequence, Text

class Error(Exception): ...

def register(
    name: Text, klass: Callable[[], BaseBrowser] | None, instance: BaseBrowser | None = ..., update_tryorder: int = ...
) -> None: ...
def get(using: Text | None = ...) -> BaseBrowser: ...
def open(url: Text, new: int = ..., autoraise: bool = ...) -> bool: ...
def open_new(url: Text) -> bool: ...
def open_new_tab(url: Text) -> bool: ...

class BaseBrowser:
    args: list[str]
    name: str
    basename: str
    def __init__(self, name: Text = ...) -> None: ...
    def open(self, url: Text, new: int = ..., autoraise: bool = ...) -> bool: ...
    def open_new(self, url: Text) -> bool: ...
    def open_new_tab(self, url: Text) -> bool: ...

class GenericBrowser(BaseBrowser):
    args: list[str]
    name: str
    basename: str
    def __init__(self, name: Text | Sequence[Text]) -> None: ...
    def open(self, url: Text, new: int = ..., autoraise: bool = ...) -> bool: ...

class BackgroundBrowser(GenericBrowser):
    def open(self, url: Text, new: int = ..., autoraise: bool = ...) -> bool: ...

class UnixBrowser(BaseBrowser):
    raise_opts: list[str] | None
    background: bool
    redirect_stdout: bool
    remote_args: list[str]
    remote_action: str
    remote_action_newwin: str
    remote_action_newtab: str
    def open(self, url: Text, new: int = ..., autoraise: bool = ...) -> bool: ...

class Mozilla(UnixBrowser):
    remote_args: list[str]
    remote_action: str
    remote_action_newwin: str
    remote_action_newtab: str
    background: bool

class Galeon(UnixBrowser):
    raise_opts: list[str]
    remote_args: list[str]
    remote_action: str
    remote_action_newwin: str
    background: bool

class Chrome(UnixBrowser):
    remote_args: list[str]
    remote_action: str
    remote_action_newwin: str
    remote_action_newtab: str
    background: bool

class Opera(UnixBrowser):
    remote_args: list[str]
    remote_action: str
    remote_action_newwin: str
    remote_action_newtab: str
    background: bool

class Elinks(UnixBrowser):
    remote_args: list[str]
    remote_action: str
    remote_action_newwin: str
    remote_action_newtab: str
    background: bool
    redirect_stdout: bool

class Konqueror(BaseBrowser):
    def open(self, url: Text, new: int = ..., autoraise: bool = ...) -> bool: ...

class Grail(BaseBrowser):
    def open(self, url: Text, new: int = ..., autoraise: bool = ...) -> bool: ...

if sys.platform == "win32":
    class WindowsDefault(BaseBrowser):
        def open(self, url: Text, new: int = ..., autoraise: bool = ...) -> bool: ...

if sys.platform == "darwin":
    class MacOSX(BaseBrowser):
        name: str
        def __init__(self, name: Text) -> None: ...
        def open(self, url: Text, new: int = ..., autoraise: bool = ...) -> bool: ...

    class MacOSXOSAScript(BaseBrowser):
        def __init__(self, name: Text) -> None: ...
        def open(self, url: Text, new: int = ..., autoraise: bool = ...) -> bool: ...



import ast
from typing import Any

class StrAndRepr: ...

def evaluate_string(string: str | ast.AST) -> Any: ...

class Token(StrAndRepr):
    type: str
    def __init__(self, type: str) -> None: ...



from typing import Iterable

PREFIXES: list[str]
ENABLE_USER_SITE: bool | None
USER_SITE: str | None
USER_BASE: str | None

def main() -> None: ...
def addsitedir(sitedir: str, known_paths: Iterable[str] | None = ...) -> None: ...
def getsitepackages(prefixes: Iterable[str] | None = ...) -> list[str]: ...
def getuserbase() -> str: ...
def getusersitepackages() -> str: ...



from braintree.resource import Resource as Resource
from braintree.util.datetime_parser import parse_datetime as parse_datetime

class AchMandate(Resource):
    def __init__(self, gateway, attributes) -> None: ...
